;;; GAMS.EL --- Major mode for editing and viewing GAMS program files.

;; Copyright (C) 2001 Shiro Takeda

;; Version: 1.1
;; $Id: gams.el,v 1.1.1.1 2002/01/06 00:14:30 ged9203 Exp ged9203 $
;; Time-stamp: <2002-01-20 11:24:11 Shiro Takeda>
;; Time stamp: <2001-12-24 22:56:55 Shiro Takeda>

;; Author: Shiro Takeda <zbc08106@park.zero.ad.jp>
;; Maintainer: Shiro Takeda <zbc08106@park.zero.ad.jp>
;; First Created: Sun Aug 19, 2001 12:48 PM

;; This file is not part of any Emacs.

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; A copy of the GNU General Public License can be obtained from this
;; program's author (send electronic mail to
;; zbc08106@park.zero.ad.jp) or from the Free Software Foundation,
;; Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:
;;
;;
;; 			See README file!
;;
;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;	Code starts here.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'easymenu)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;	Define variables.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst gams-mode-version "1.1"
  "Version of GAMS mode.")

;;;;; Define groups.

(defgroup gams nil
  "GAMS mode for Emacs."
  :group 'applications)

(defgroup gams-faces nil
  "Faces for GAMS mode."
  :group 'gams
  :group 'faces)

(defgroup gams-keys nil
  "Keybinding for GAMS mode."
  :group 'gams
  :group 'keyboard)

;;;;; Customizable variables start here.

;;; Variables for GAMS mode.

;; from epop.el
(defcustom gams:process-command-name "gams"
  "*GAMS program file name.
If you does not include the GAMS system directory in PATH variable, you
must set the full path to GAMS in this variable like
c:/GAMS20.0/gams.exe."
  :type 'file
  :group 'gams)

(defcustom gams:process-command-option "ll=0 lo=3 pw=100 ps=9999"
  "*Command line options passed to GAMS"
  :type 'string
  :group 'gams)

(defcustom gams-statement-file "~/.gams-statement"
  "*The name of the file in which user specific statements are stored."
  :type 'file
  :group 'gams)

;; load gams-statement-file. 
(if (file-exists-p gams-statement-file)
    (load-file gams-statement-file))


(defcustom gams-statement-upcase nil
  "*Non-nil means that statement is inserted in upper case.
If you want to use lower case, set nil to this variable."
  :type 'boolean
  :group 'gams)

(defcustom gams-dollar-control-upcase nil
  "*Non-nil means that dollar control option is inserted in upper case.
If you want to use lower case, set nil to this variable."
  :type 'boolean
  :group 'gams)

(defcustom gams-use-mpsge nil
  "*If you use MPSGE, set this variable to non-nil.
Non-nil enables MPSGE statement completion and colorization."
  :type 'boolean
  :group 'gams)

(defcustom gams-fill-column 74
  "*The column number used for fill-paragraph and auto-fill-mode."
  :type 'number
  :group 'gams)

(defcustom gams-recenter-font-lock t
"Non-nil means that font-lock-fontify buffer when recentering.
If your computer is slow, you may better set this to nil."
  :type 'boolean
  :group 'gams)

(defcustom gams-file-extension '("gms")
  "*List of gams program file extensions.
If you open a file with an extension included in this list, GAMS mode
starts automatically.  It doen't matter whether upper case or lower
case."
  :type '(repeat (string :tag "value"))
  :group 'gams)

(defcustom gams-multi-process t
  "*Non-nil enables multiple GAMS processes.
Non-nil means that you can run multiple GAMS processes at the same time
in an Emacs.  If you rarely run multiple processes, you had better set it
to nil."
  :type 'boolean
  :group 'gams)
 
(defvar gams-mode-hook  nil
  "*Hook run when gams-mode starts.")

;;; from yatex.el
(defcustom gams-close-paren-always t
  "*Non-nil means that close parenthesis."
  :type 'boolean
  :group 'gams)

(defcustom gams-close-double-quotation-always t
  "*Non-nil means that close double quotation."
  :type 'boolean
  :group 'gams)

(defcustom gams-close-single-quotation-always t
  "*Non-nil means that close quotation."
  :type 'boolean
  :group 'gams)

(defcustom gams-statement-name "set"
  "*The initial value of statement."
  :type 'string
  :group 'gams)

(defcustom gams-dollar-control-name "title"
  "*The initial value of dollar control."
  :type 'string
  :group 'gams)
  
(defcustom gams-user-comment
  "
*------------------------------------------------------------------------	
* %       
*------------------------------------------------------------------------
"
  "*User defined comment template.
You can insert the comment template defined in this variable by executing
`gams-insert-comment'.  % in the string indicates the cursor place and will
disappear after template insertion.
NB: You cannot include double quoatations and backslashes in this variables!"
  :type 'string
  :group 'gams)

;; from epolib.el
(defcustom gams-default-pop-window-height 10
  "The default GAMS process buffer height.
If integer, sets the window-height of process buffer.  If string, sets the
percentage of it.  If nil, use default pop-to-buffer."
  :type 'number
  :group 'gams)

(defcustom gams-docs-dir "c:/GAMS20.0/docs"
      "The directory of GAMS PDF manuals.

If the GAMS system directory is c:/GAMS20.0, then set c:/GAMS20.0/docs in
this variable."
  :type 'directory
  :group 'gams)

(defcustom gams-docs-view-program "c:/Program Files/Adobe/Acrobat 4.0/Reader/AcroRd32.exe"
  "The full path to the pdf file viewer."
  :type 'file
  :group 'gams)

(defvar gams-manuals-alist
      '(
	("User-Manual" . "GAMSUsersGuide")
	("Tutorial" . "Tutorial")
	("PC-Install" . "pc-install")
	("Unix-Install" . "unix-install")
	("Solver-BDMLP" . "bdmlp")
	("Solver-CONOPT" . "conopt")
	("Solver-CPLEX" . "cplex")
	("Solver-DECIS" . "decis")
	("Solver-DICOPT" . "dicopt")
	("Solver-GAMSBAS" . "gamsbas")
	("Solver-GAMSCHK" . "gamschk")
	("Solver-MILES" . "miles")
	("Solver-MINOS" . "minos")
	("Solver-MPSGE" . "mpsge")
	("Solver-MPSWRITE" . "mpswrite")
	("Solver-OSL" . "osl")
	("Solver-OSLSE" . "oslse")
	("Solver-PATH" . "path")
	("Solver-SBB" . "sbb")
	("Solver-SNOPT" . "snopt")
	("Solver-SOLVERINTRO" . "solverintro")
	("Solver-XA" . "xa")
	("Solver-XPRESS" . "xpress")
	("Solver-XPRESSLICENSING" . "xpresslicensing"))
      "Alist of GAMS manuals.
(label . pdf-file-name)
")

;;; Variables for GAMS-LST mode.

(defvar gams-lst-mode-hook  nil
  "*GAMS-LST mode hooks.")


(defvar gams-lst-gms-extention  "gms"
  "*GAMS program file extention.")

;;; Key for GAMS-LST mode.

(defcustom gams-lk-1 "i"
  "Key for `gams-lst-jump-to-input-file'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-lk-2 "u"
  "Key for `gams-lst-jump-to-error-file'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-lk-3 "y"
  "Key for `gams-lst-view-error'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-lk-4 "b"
  "Key for `gams-lst-jump-to-input-file-2'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-lk-5 "L"
  "Key for `gams-lst-jump-to-line'."
  :type 'string
  :group 'gams-keys)


;;; Other customizable variables.

(defvar gams-statement-up
      '("SET" "SETS" "SCALAR" "SCALARS" "TABLE" "PARAMETER" "PARAMETERS"
	"EQUATION" "EQUATIONS" "VARIABLE" "VARIABLES"
	"POSITIVE VARIABLE" "POSITIVE VARIABLES"
	"NEGATIVE VARIABLE" "NEGATIVE VARIABLES"
	"INTEGER VARIABLE" "INTEGER VARIABLES"
	"BINARY VARIABLE" "BINARY VARIABLES"
	"ALIAS"
	"OPTION"
	"SOLVE" "MODEL" "DISPLAY" "LOOP" "IF" "SUM" "PROD")
      "The default list of GAMS statements.  Used for completion inserting")

(defvar gams-dollar-control-up
  '("BATINCLUDE" "EXIT" "INCLUDE" "LIBINCLUDE"
    "OFFTEXT" "ONTEXT" "SETGLOBAL" "SYSINCLUDE"
    "TITLE")
  "The default list of GAMS dollar control options.")

(defvar gams-statement-mpsge
    ; MPSGE
  '("MODEL:" "COMMODITIES:" "CONSUMERS:" "SECTORS:" "PROD:"
    "DEMAND:" "REPORT:" "CONSTRAINT:" "AUXILIARY:")
  "The default list of MPSGE statements.")


(defvar gams-statement-regexp-default
  '("ABORT" "ACRONYM" "ACRONYMS" "ALIAS" "ALL"
;    "AND" "ASSIGN" "CARD"
    "DIAG" "DISPLAY" "EPS"
;    "EQ"
;    "GE" "GT"
    "INF" "LE" "LOOP"
;    "LT"
    "MAXIMAZING"
    "MINIMIZING" "MODEL[S]?" "NA" "NE"
;    "NOT"
    "OPTION[S]?"
;    "OR"
    "ORD" "PROD"
    "SAMEAS" "SCALAR[S]?" "SET[S]?" "SMAX" "SMIN" "SOS1" "SOS2" "SUM" "SYSTEM"
    "TABLE[S]?" "USING" "XOR" "YES" "REPEAT" "UNTIL" "WHILE" "IF" "THEN" "ELSE"
    "ELSEIF"
    "SEMICONT" "SEMIINT" "FILE" "FILES" "PUTPAGE" "PUTTL" "FREE" "NO" "SOLVE"
    "FOR" "PARAMETER[S]?" "EQUATION[S]?"
    "\\(POSITIVE\\|NEGATIVE\\|BINARY\\|INTEGER\\)[ \t]+VARIABLE[S]?" "VARIABLE[S]?"
    )
  "*The list of GAMS statements (reserved words) for font-lock (coloring).
These statements are colorized only if they lie at the beginning of a line.")


;;;;; Customizable variables end here.


;;;;; Non-customizable variables.

(defvar gams-run-key ?s
  "*GAMS run key")

(defvar gams-kill-key ?k
  "*GAMS kill key")

(defvar gams-win32
  (memq system-type '(ms-dos windows-nt)))
(defvar gams-dos (memq system-type '(ms-dos windows-nt OS/2)))
(defvar gams-emacs-19 (= (string-to-int emacs-version) 19))
(defvar gams-emacs-20 (= (string-to-int emacs-version) 20))
(defvar gams-emacs-21 (= (string-to-int emacs-version) 21))

(defvar gams-simple-messages nil
  "Non-nil makes minibuffer messages simpler.")

(defvar gams-comment-prefix "*"
  "*GAMS comment prefix.")

(defvar gams-fill-prefix nil
  "*fill-prefix used for auto-fill-mode.
The default value is nil.")

(defvar gams-lst-extention "lst"
  "GAMS LST file extention.")

(defvar gams-fill-prefix nil
  "*fill-prefix used for auto-fill-mode.
The default value is nil.")

(defun gams-list-upcase (list)
  "Make each element of a LIST to upper case
and return a new list."
  (let* ((temp-list list)
	 list-new ele)
    (while temp-list
      (setq ele (upcase (car temp-list)))
      (setq list-new (cons ele list-new))
      (setq temp-list (cdr temp-list)))
    list-new))

(defun gams-list-downcase (list)
  "Make each element of a LIST to lower case
and return a new list."
  (let* ((temp-list list)
	 list-new ele)
    (while temp-list
      (setq ele (downcase (car temp-list)))
      (setq list-new (cons ele list-new))
      (setq temp-list (cdr temp-list)))
    list-new))

(defvar gams-user-statement-list nil)
(defvar gams-user-dollar-control-list nil)
(defvar gams-paragraph-start "[ \t]*$\\|^[\f\n]")
(defvar gams*command-process-buffer "*GAMS")
(defvar gams-statement-down
  (gams-list-downcase gams-statement-up))
(defvar gams-dollar-control-down
  (gams-list-downcase gams-dollar-control-up))
(defvar gams-statement-alist nil)
(defvar gams-dollar-control-alist nil)
(defvar gams-statement-regexp nil)

;(defvar gams:current-process-buffer nil)

;;; From EPO. 
(defconst gams:frame-feature-p
  (and (fboundp 'make-frame) window-system))

;;; I don't understand the meaning of this variable.
(defvar gams:shell-c
  (or (and (boundp 'shell-command-switch) shell-command-switch)
      (and (boundp 'shell-command-option) shell-command-option)
      (and (string-match "command.com\\|cmd.exe\\|start.exe" shell-file-name)
	   "/c")
      "-c")
  "*Command option for shell")

(if (fboundp 'buffer-substring-no-properties)
    (fset 'gams*buffer-substring 'buffer-substring-no-properties)
  (fset 'gams*buffer-substring 'buffer-substring))

(cond
 ((fboundp 'screen-height)
  (fset 'gams*screen-height 'screen-height)
  (fset 'gams*screen-width 'screen-width))
 ((fboundp 'frame-height)
  (fset 'gams*screen-height 'frame-height)
  (fset 'gams*screen-width 'frame-width))
 (t (error "I don't know how to run GAMS on this Emacs...")))


;;; Autoload setting.

; Funciotn for creating regexp.

(defun gams-create-regexp (list)
  "Create the regular expression from a LIST."
  (let ((temp-list list)
	new-list)
    (while temp-list
      (setq new-list (concat "\\|" (car temp-list) new-list))
      (setq temp-list (cdr temp-list)))
    new-list))

; For GAMS mode.
(setq auto-mode-alist
	(cons
	 (cons
	  (format "\\.\\(xyz%s\\)$" (gams-create-regexp
				  (append (gams-list-downcase gams-file-extension)
					  (gams-list-upcase gams-file-extension))))
	  'gams-mode) auto-mode-alist))
(autoload 'gams-mode "gams" "Enter GAMS mode" t)

; For GAMS-LST mode.
(setq auto-mode-alist
	(cons (cons "\\.\\(LST\\|lst\\)$" 'gams-lst-mode) auto-mode-alist))
(autoload 'gams-lst-mode "gams" "Enter GAMS-LST mode" t)

;; Load the values of various variables.
;(require 'gams-var)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;	Code for font-lock.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar gams-mpsge-face 'gams-mpsge-face
  "Face for MPSGE statements.")
(defvar gams-misc-face	'gams-misc-face
  "Face for misc.")
(defvar gams-comment-face 'gams-comment-face
  "Face for comment.")
(defvar gams-dollar-face 'gams-dollar-face
  "Face for dollar control options.")
(defvar gams-statement-face 'gams-statement-face
  "Face for GAMS statments.")

(defvar gams-lst-par-face 'gams-lst-par-face
  "Face for PARAMETER in GAMS-LST mode.")
(defvar gams-lst-set-face 'gams-lst-set-face
  "Face for PARAMETER in GAMS-LST mode.")
(defvar gams-lst-var-face 'gams-lst-var-face
  "Face for VAR in GAMS-LST mode.")
(defvar gams-lst-equ-face 'gams-lst-equ-face
  "Face for EQU in GAMS-LST mode.")
(defvar gams-lst-vri-face 'gams-lst-vri-face
  "Face for VARIABEL in GAMS-LST mode.")
(defvar gams-lst-oth-face 'gams-lst-oth-face
  "Face for misc in GAMS-LST mode.")
(defvar gams-lst-warning-face 'gams-lst-warning-face
  "Face for warning in GAMS-LST mode.")
(defvar gams-lst-program-face 'gams-lst-program-face
  "Face for program listing in GAMS-LST mode.")

(defvar gams-ol-loo-face 'gams-ol-loo-face
  "Face for program listing in GAMS-LST mode.")


;; But now we do it the custom way.  Note that `defface' will not overwrite any
;; faces declared above via `custom-declare-face'.

;;; GAMS mode.

(defface gams-comment-face
    '((((class color) (background light)) (:bold t :foreground "limegreen"))
      (((class color) (background dark)) (:bold t :foreground "lawn green")))
  ""
  :group 'gams-faces)

(defface gams-mpsge-face
    '((((class color) (background light)) (:bold t :foreground "gold"))
      (((class color) (background dark)) (:bold t :foreground "hot pink")))
  ""
  :group 'gams-faces)

(defface gams-statement-face
  '((((class color) (background light)) (:bold t :foreground "DodgerBlue"))
    (((class color) (background dark)) (:bold t :foreground "cyan")))
  ""
  :group 'gams-faces)

(defface gams-dollar-face
  '((((class color) (background light)) (:bold t :foreground "hot pink"))
    (((class color) (background dark)) (:bold t :foreground "yellow")))
  ""
  :group 'gams-faces)

(defface gams-misc-face
  '((((class color) (background light)) (:foreground "red" :bold t))
    (((class color) (background dark)) (:foreground "red" :bold t)))
  ""
  :group 'gams-faces)


;;; GAMS-LST mode.

(defface gams-lst-par-face
  '((((class color) (background light)) (:bold t :foreground "DodgerBlue"))
    (((class color) (background dark)) (:bold t :foreground "yellow")))
  ""
  :group 'gams-faces)

(defface gams-lst-set-face
  '((((class color) (background light)) (:bold t :foreground "light blue"))
    (((class color) (background dark)) (:bold t :foreground "sandy brown")))
  ""
  :group 'gams-faces)

(defface gams-lst-var-face
  '((((class color) (background light)) (:bold t :foreground "hot pink"))
    (((class color) (background dark)) (:bold t :foreground "cyan")))
  ""
  :group 'gams-faces)

(defface gams-lst-equ-face
  '((((class color) (background light)) (:bold t :foreground "lawn green"))
    (((class color) (background dark)) (:bold t :foreground "pink")))
  ""
  :group 'gams-faces)

(defface gams-lst-vri-face
    '((((class color) (background light)) (:bold t :foreground "purple"))
      (((class color) (background dark)) (:bold t :foreground "pale green")))
  ""
  :group 'gams-faces)

(defface gams-lst-oth-face
  '((((class color) (background light)) (:bold t :foreground "gray60"))
    (((class color) (background dark)) (:bold t :foreground "cornsilk")))
  ""
  :group 'gams-faces)

(defface gams-lst-warning-face
  '((((class color) (background light)) (:bold t :foreground "red"))
    (((class color) (background dark)) (:bold t :foreground "red")))
  ""
  :group 'gams-faces)

(defface gams-lst-program-face
  '((((class color) (background light)) (:foreground "goldenrod"))
    (((class color) (background dark)) (:foreground "khaki")))
  ""
  :group 'gams-faces)


;;;

(defface gams-ol-loo-face
  '((((class color) (background light)) (:bold t :foreground "maroon"))
    (((class color) (background dark)) (:bold t :foreground "#7777ff")))
  ""
  :group 'gams-faces)


;;;

(defvar gams-font-lock-keywords
      (list
       ; MPSGE syntax.
       (list "\\($model\\|$MODEL\\|$commodities\\|$COMMODITIES\\|$consumer[s]?\\|$CONSUMERS\\|$sectors\\|$SECTORS\\|$prod\\|$PROD\\|$demand\\|$DEMAND\\|$report\\|$REPORT\\|$CONSTRAINT\\|$constraint\\|$auxiliary\\|$AUXILIARY\\|$funlog\\|$FUNLOG\\|$echop\\|$ECHOP\\|$peps\\|$PEPS\\|$eulchk\\|$EULCHK\\|$walchk\\|$WALCHK\\)"
		 '(0 gams-mpsge-face))
       ; ontext - offtext pair.
       (if gams-emacs-21
	   (list "\\(^$ontext\\|^$ONTEXT\\)\\([^$]\\|$[^o]\\|$o[^f]\\|$of[^f]\\|$off[^t]\\|$offt[^e]\\|$offte[^x]\\|$offtex[^t]\\)+\\(^$offtext\\|^$OFFTEXT\\)"
	       '(0 gams-comment-face))
	 (list "\\(^$ontext\\|^$ONTEXT\\)[^$]*\\(^$offtext\\|^$OFFTEXT\\)"
	       '(0 gams-comment-face)))
       ; Commented out text by *
       (list "^\\(*\\).*$" '(0 gams-comment-face))
       ; dollar control option.
       (list "[ \n\t]+\$[a-zA-Z]+[ \t\n]"
	     '(0 gams-dollar-face))
       (list "^\$[a-zA-Z]+[ \t\n]"
	     '(0 gams-dollar-face))
       ; semicolon
       (list ";" '(0 gams-misc-face))
       ; Standard GAMS statement.
       '(gams-statement-regexp-func
	 . gams-statement-face)
       ; Commented out text by !
       (list "\\([!]\\).*$" '(0 gams-comment-face))
       ))

(defvar gams-lst-font-lock-keywords
  (list
    (list "^\\*\\*\\*\\*[^\n]+" '(0 gams-lst-warning-face))
    (list "^----[ ]+[0-9]+[ ]PARAMETER[ ]+" '(0 gams-lst-par-face))
    (list "^[ ]+\\(PARAMETER\\)[ ]+" '(1 gams-lst-par-face))
    (list "^----[ ]+[0-9]+[ ]SET[ ]+" '(0 gams-lst-set-face))
    (list "^----[ ]+[0-9]+[ ]VARIABLE[ ]+" '(0 gams-lst-vri-face))
    (list "^[ ]+\\(VARIABLE\\)[ ]+" '(1 gams-lst-vri-face))
    (list "\\(----[ ]+VAR[ ]+[^ ]+\\)[ ]*[^\n]+" '(1 gams-lst-var-face))
    (list "\\(----[ ]+EQU[ ]+[^ ]+\\)[ ]*[^\n]+" '(1 gams-lst-equ-face))
    ;; Copy of program file
    (list "^[ ]?[ ]?[ ]?[ ]?[ ]?[ ]?\\([0-9]+[ ][ ].*\\)" '(0 gams-lst-program-face))
    ))

(defvar gams-ol-font-lock-keywords
  (list
    (list "^[ ]+\\(PAR\\)" '(1 gams-lst-par-face))
    (list "^[ ]+\\(SET\\)" '(1 gams-lst-set-face))
    (list "^[ ]+\\(SUM\\)" '(1 gams-lst-warning-face))
    (list "^[ ]+\\(VAR\\)" '(1 gams-lst-var-face))
    (list "^[ ]+\\(VRI\\)" '(1 gams-lst-vri-face))
    (list "^[ ]+\\(EQU\\)" '(1 gams-lst-equ-face))
    (list "^[ ]+\\(LOO\\)" '(1 gams-ol-loo-face))
    (list "^[ ]+\\(OTH\\)" '(1 gams-lst-oth-face))
    (list "^\\(COM: \\).*" '(0 gams-comment-face))
    (list "^[*] \\(.*\\)" '(0 gams-mpsge-face))
    ))

;(load "gams-font-lock.el")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;	Code for GAMS mode.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun gams-insert-tab ()
  "Insert a tab."
  (interactive)
  (insert "	"))

;; Key assignment.
(defvar gams-mode-map (make-keymap) "keymap for gams-mode")
(let ((map gams-mode-map))
  (define-key map "\C-ck" 'gams-insert-statement)
  (define-key map "\C-cd" 'gams-insert-dollar-control)
  (define-key map "\C-cv" 'gams-view-lst)
  (define-key map "\C-cj" 'gams-jump-to-lst)    
  (define-key map "\C-ct" 'gams-start-menu)
  (define-key map "\C-ce" 'gams-template)

  (define-key map "(" 'gams-insert-parens)
  (define-key map "\"" 'gams-insert-double-quotation)
  (define-key map "`" 'gams-insert-single-quotation)

  (define-key map "\C-l" 'gams-recenter)
  (define-key map "\C-co" 'gams-insert-comment)

  (define-key map "$" 'gams-insert-dollar-control)
  (define-key map "\t" 'gams-insert-tab)

  (define-key map "\C-cg" 'gams-jump-on-off-text)
  (define-key map "\C-cm" 'gams-view-docs)  
  )

;;; Menu for GAMS mode.
(easy-menu-define 
  gams-menu gams-mode-map "Menu keymap for GAMS mode."
  '("GAMS"
    ["Insert GAMS statement" gams-insert-statement t]
    ["Insert GAMS dollar control" gams-insert-dollar-control t]
    ["Switch to the LST file and show error" gams-view-lst t]
    ["Switch to the LST file" gams-jump-to-lst t]
    ["Evoke the TEMPLATE mode" gams-template t]
    ("Process"
     ["Start GAMS" (gams-start-menu nil ?s) t]
     ["Kill GAMS process" (gams-start-menu nil ?k) t]
     ["Process menu" (gams-start-menu) t]
     )
    "--"
    ["Insert parenthesis" gams-insert-parens t]
    ["Insert double quotations" gams-insert-double-quotation t]
    ["Insert single quotations" gams-insert-single-quotation t]
    "--"
    ["Recentering" gams-recenter t]
    ["Insert a comment template" gams-insert-comment t]
    ["Jump between ontext and offtext" gams-jump-on-off-text t]
    ["View GAMS pdf manuals." gams-view-docs t]
    ))

;;; 
(defun gams-mode ()
  "Major mode for editing GAMS program file.

The following commands are available in the GAMS-LST mode:

\\[gams-insert-statement]		Insert GAMS statement with completion.
\\[gams-insert-dollar-control]		Insert GAMS statement (dollar control option).
\\[gams-view-lst]		Switch to the LST file and show errors if exist.
\\[gams-jump-to-lst]		Switch to the LST file.
\\[gams-start-menu]		Run GAMS on a file you are editing or Kill GAMS process.
\\[gams-template]		Evoke the TEMPLATE mode.

\\[gams-insert-parens]		Insert parenthesis.
\\[gams-insert-double-quotation]		Insert double quotations.
\\[gams-insert-single-quotation]		Insert single quotations.

\\[gams-recenter]		Recenter.
\\[gams-insert-comment]		Insert comment template.
\\[gams-jump-on-off-text]		Jump between ontext and offtext.
\\[gams-view-docs]		View GAMS pdf manuals.
"
  
  (interactive)
  (kill-all-local-variables)  
  (setq major-mode 'gams-mode)
  (setq mode-name "GAMS")
  (use-local-map gams-mode-map)
;  (easy-menu-add todo-menu)
;  (setq paragraph-separate "\*/\*")
  (setq fill-prefix "\t\t")
;  (setq outline-regexp "\\*/\\*")
;  (outline-minor-mode 1)
;  (hide-other)
;  (auto-fill-mode 1)

; mapcar is a built-in function.

; Apply FUNCTION to each element of SEQUENCE, and make a list of the results.
; The result is a list just as long as SEQUENCE.
; SEQUENCE may be a list, a vector, a bool-vector, or a string.
; (mapcar FUNCTION SEQUENCE)
  (mapcar
   'make-local-variable
   '(fill-column
     fill-prefix
     paragraph-start
;      paragraph-separate
     indent-line-function
     comment-start
     comment-start-skip))
  (setq
   fill-column gams-fill-column
   fill-prefix gams-fill-prefix
   paragraph-start gams-paragraph-start
;    paragraph-separate gams-paragraph-separate
   indent-line-function 'gams-indent-line-function
   comment-start gams-comment-prefix
   comment-end ""
   comment-start-skip "^*+[ \t]*")

  ;; Create the alist of statements.  Is this necessary?  See
  ;; `gams-statement-update'.
  (setq gams-statement-alist
	(gams-statement-to-alist gams-statement-up gams-statement-upcase))
  
  ;; Create the alist of dollar control options.  Is this necessary?  See
  ;; `gams-statement-update'.
  (if gams-use-mpsge
      ;; Use mpsge.
      (progn
	(setq gams-dollar-control-alist
	      (gams-statement-to-alist
	       (append gams-dollar-control-up gams-statement-mpsge)
	       gams-dollar-control-upcase)))
    ;; Not use mpsge
    (setq gams-dollar-control-alist
	  (gams-statement-to-alist gams-dollar-control-up gams-dollar-control-upcase)))

  ;; Font-lock
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(gams-font-lock-keywords t t))

  ;; Local variables to store window configurations.
  (make-local-variable 'gams-temp-window)
    
  ;; Update.
  (gams-statement-update)

  ;; menu
;  (gams-setup-menu)
  (easy-menu-add gams-menu)

  ; Run hook
  (run-hooks 'gams-mode-hook))


;;; gams-mode ends.


;;; Functions for indentation.  From indent.el
(defun gams-indent-line-function ()
  "Indent a new line like previous nonblank line."
  (interactive)
  (gams-indent-function  t))

(defun gams-indent-function (&optional unindented-ok)
  "Indentation function in GAMS mode."
  (interactive "P")
  (if (and abbrev-mode
	   (eq (char-syntax (preceding-char)) ?w))
      (expand-abbrev))
  (let ((start-column (current-column))
	(point-here (point))
	point-0	point-a	flag indent)
    (beginning-of-line)
    (setq point-0 (point))
    (if (re-search-backward ";\n" nil t)
	(progn
	  (goto-char (setq point-a (match-end 0)))
	  (while (< point-a point-here)
	    (if (looking-at "[^\n ]")
		(setq flag t))
	    (setq point-a (+ point-a 1))
	    (goto-char point-a))
	  (if (null flag)
	      nil
	    
	    (save-excursion
	      (goto-char point-0)
	      (if (re-search-backward "^[^\n]" nil t)
		  (let ((end (save-excursion (forward-line 1) (point))))
		    (move-to-column start-column)
		    ;; Is start-column inside a tab on this line?
		    (if (> (current-column) start-column)
			(backward-char 1))
		    (or (looking-at "[ \t]")
			unindented-ok
			(skip-chars-forward "^ \t" end))
		    (skip-chars-forward " \t" end)
		    (or (= (point) end) (setq indent (current-column))))))
	    (goto-char point-0)
	    (if indent
		(let ((opoint (point-marker)))
		  (delete-region (point) (progn (skip-chars-backward " \t") (point)))
		  (indent-to indent 0)
		  (if (> opoint (point))
		      (goto-char opoint))
		  (move-marker opoint nil))
	      (tab-to-tab-stop)))))))


(defun gams-statement-to-alist (list flag)
  "Transform a LIST to an alist.
Non-nil of FLAG means use upper case."
    (let ((list-name list)
	  gams-alist)
      (if (null flag)
	  (setq list-name (gams-list-downcase list-name))
	nil)
      (while list-name
	(setq gams-alist (cons (list (car list-name)) gams-alist))
	(setq list-name (cdr list-name)))
      ;; Return the alist.
      gams-alist))

(defun gams-statement-update ()
  "Update gams-statement-alist and gams-dollar-control-alist."
  ;; Update `gams-statement-alist'.
  (setq gams-statement-alist
	(gams-statement-to-alist
	 (append gams-statement-up gams-user-statement-list) gams-statement-upcase))
  ;; Update `gams-dollar-control-alist'.
  (setq gams-dollar-control-alist
	(gams-statement-to-alist
	 ;; If you use MPSGE
	 (if gams-use-mpsge		
	     (append gams-dollar-control-up gams-statement-mpsge
		     gams-user-dollar-control-list)
	   (append gams-dollar-control-up gams-user-dollar-control-list))
	 gams-dollar-control-upcase)))

;;; From yatexlib.el

;;;###autoload
(defun gams-minibuffer-complete ()
  "Complete in minibuffer.
  If the symbol 'delim is bound and is string, its value is assumed to be
the character class of delimiters.  Completion will be performed on
the last field separated by those delimiters.
  If the symbol 'quick is bound and is 't, when the try-completion results
in t, exit minibuffer immediately."
  (interactive)
  (save-restriction
    (narrow-to-region
     (if (fboundp 'field-beginning) (field-beginning (point-max)) (point-min))
     (point-max))
    (let ((md (match-data)) beg word compl
	  (quick (and (boundp 'quick) (eq quick t)))
	  (displist ;function to display completion-list
	   (function
	    (lambda ()
	      (with-output-to-temp-buffer "*Completions*"
		(display-completion-list
		 (all-completions word minibuffer-completion-table)))))))
      (setq beg (if (and (boundp 'delim) (stringp delim))
		    (save-excursion
		      (skip-chars-backward (concat "^" delim))
		      (point))
		  (point-min))
	    word (buffer-substring beg (point-max))
	    compl (try-completion word minibuffer-completion-table))
      (cond
       ((eq compl t)
	(if quick (exit-minibuffer)
	  (let ((p (point)) (max (point-max)))
	    (unwind-protect
		(progn
		  (goto-char max)
		  (insert " [Sole completion]")
		  (goto-char p)
		  (sit-for 1))
	      (delete-region max (point-max))
	      (goto-char p)))))
       ((eq compl nil)
	(ding)
	(save-excursion
	  (let (p)
	    (unwind-protect
		(progn
		  (goto-char (setq p (point-max)))
		  (insert " [No match]")
		  (goto-char p)
		  (sit-for 2))
	      (delete-region p (point-max))))))
       ((string= compl word)
	(funcall displist))
       (t (delete-region beg (point-max))
	  (insert compl)
	  (if quick
	      (if (eq (try-completion compl minibuffer-completion-table) t)
		  (exit-minibuffer)
		(funcall displist)))))
      (store-match-data md))))

;;; From yatex.el

(defvar gams-statement-completion-map nil
  "*Key map used at gams completion of statements in the minibuffer.")
(if gams-statement-completion-map nil
   (setq gams-statement-completion-map
	(copy-keymap (or (and (boundp 'gmhist-completion-map)
			      gmhist-completion-map)
			 minibuffer-local-completion-map)))
  (define-key gams-statement-completion-map
    " " 'gams-minibuffer-complete)
  (define-key gams-statement-completion-map
    "\C-i" 'gams-minibuffer-complete))

;;; ???
(defvar gams-read-statement-history nil "Holds history of statement.")
(put 'gams-read-statement-history 'no-default t)
(defun gams-read-statement (prompt &optional predicate initial)
  "Read a GAMS statements with completion."
;  (YaTeX-sync-local-table 'tmp-section-table)
  (let ((minibuffer-completion-table
	 (append gams-alist)))
;	 (append tmp-section-table user-section-table section-table)))
    (read-from-minibuffer
     prompt initial gams-statement-completion-map nil
     'gams-read-statement-history)))

(defun gams-register (name &optional flag)
  "Register a statement or a dollar-control.

NAME is the name of a statement or dollar-control registered.  If FLAG is
non-nil, it is a dollar-control."
  (interactive)
  (let* ((curr-buff (current-buffer))
	 (temp-buff " *gams-register*")
	 (temp-file gams-statement-file)
	 (temp-list (if flag
			gams-user-dollar-control-list
		      gams-user-statement-list))
	 (old-list temp-list)
	 (list-name
	  (if flag
	      "gams-user-dollar-control-list"
	    "gams-user-statement-list"))
	 new-list)
    (save-excursion
      ;; Make a new list.
      (setq new-list (append (list name) temp-list))
      ;; Switch to the temporary buffer.
      (get-buffer-create temp-buff)
      (set-buffer temp-buff)
      (erase-buffer)
      ;; Write the content of the list.
      (gams-stat-write-list new-list flag)
      ;; Check whether the variable is defined correctly.
      (eval-buffer)
      ;; Store the content of buffer
      (setq temp-cont (buffer-substring (point-min) (point-max)))
      ;; Delete the list-name part.
      (set-buffer (find-file-noselect temp-file))
      (goto-char (point-min))
      ;; Check whether the list-name part exists or not.
      (if (null (re-search-forward
		 (concat
		  "\\(setq\\) " list-name)
		 nil t))
	  ;; If it doesn't exists, do nothing.
	  nil
	;; If it exists, delete it.
	(let (point-beg point-en)
	  (goto-char (match-beginning 1))
	  (beginning-of-line)
	  (setq point-beg (point))
	  (forward-sexp 1)
	  (forward-line 1)
	  (setq point-en (point))
	  (delete-region point-beg point-en)))
      ;; Insert the content.
      (goto-char (point-min))
      (insert temp-cont)
      ;;
      (eval-buffer)
      ;; Save buffer of gams-statement-file.
      (save-buffer (get-file-buffer temp-file))
      (kill-buffer (get-file-buffer temp-file))
      ;; kill the temporary buffer.
      (kill-buffer temp-buff)
      ;; Replace the old list with the new list.
      (setq old-list new-list)
      (gams-statement-update)
      (set-buffer curr-buff)
      )))

(defun gams-stat-write-list (list &optional flag)
  "Write gams-user-statement-list or gams-user-dollar-control-list in a buffer.

If flag is non-nil, choose dollar control."
  (let ((list-name (if flag
		       "gams-user-dollar-control-list"
		     "gams-user-statement-list")))
    (erase-buffer)
    (insert-string (concat "(setq " list-name " '(\n"))
    (goto-char (point-max))
    ;; Repeat.
    (while list
      (insert (concat "\"" (car list) "\"\n"))
      (goto-char (point-max))
      (setq list (cdr list)))
    ;; Last.
    (insert "))\n")))

(defun gams-insert-statement (arg &optional beg end cmd)
  "Insert GAMS statement with completion.
List of candidates is created from elements of `gams-statement-up'
and `gams-user-statement-list'."
  
  (interactive "P")
  (gams-statement-update)
  (unwind-protect
      (let*
	  ((gams-alist gams-statement-alist)
	   (completion-ignore-case t)
	   key1
	   (source-window (selected-window))
	   guess
	   (statement
	    (or cmd
		(progn
		  (setq guess
			(if gams-statement-upcase
			    (upcase gams-statement-name)
			  (downcase gams-statement-name)))
		  (gams-read-statement
		   (if gams-simple-messages
		       (format "Statement (default %s): " guess)
		     (if (> (minibuffer-depth) 0)
			 (format "(default %s)%s: "
				 guess
				 (format "[level:%d]" (minibuffer-depth)))
		       (format "Insert statement (default = %s): " guess)))
		   nil))))
	   (statement (if (string= statement "") guess statement))
	   );;let

	(if gams-statement-upcase
	    (setq statement (upcase statement))
	  (setq statement (downcase statement)))
	(setq gams-statement-name statement)
	;; Register or not?
	(if (null (member (list statement) gams-statement-alist))
	    (progn
	      (message "Store `%s' for future use?  Type `y' if yes: " statement)
	      (setq key1 (read-char))
	      (if (equal key1 ?y)
		  (progn
		    (setq statement (upcase statement))
		    (gams-register statement))
		nil))
	  nil)
	;; Insert.
	(if gams-statement-upcase
	    (setq statement (upcase statement))
	  (setq statement (downcase statement)))
	(insert-string statement))
    (if (<= (minibuffer-depth) 0) (use-global-map global-map))
    (insert "")))	;insert dummy string to fontify(Emacs20)

(defvar gams-dollar-completion-map nil
  "*Key map used at gams completion of dollar operation in the minibuffer.")
(if gams-dollar-completion-map nil
   (setq gams-dollar-completion-map
	(copy-keymap (or (and (boundp 'gmhist-completion-map)
			      gmhist-completion-map)
			 minibuffer-local-completion-map)))
  (define-key gams-dollar-completion-map
    " " 'gams-minibuffer-complete)
  (define-key gams-dollar-completion-map
    "\C-i" 'gams-minibuffer-complete)
  (define-key gams-dollar-completion-map
    "$" 'gams-minibuffer-insert-dollar)
  (define-key gams-dollar-completion-map
    "@" 'gams-minibuffer-insert-dollar))

(defvar gams-flag-dollar nil)
(defun gams-minibuffer-insert-dollar ()
  "???"
  (interactive)
  (setq gams-flag-dollar t)
  (exit-minibuffer))

(defvar gams-read-dollar-history nil "Holds history of dollar control.")
(put 'gams-read-dollar-history 'no-default t)
(defun gams-read-dollar-control (prompt &optional predicate initial)
  "Read a GAMS dollar control operation with completion."
  (let ((minibuffer-completion-table
	 (append gams-alist)))
    (read-from-minibuffer
     prompt initial gams-dollar-completion-map nil
     'gams-read-dollar-history)))

(defun gams-insert-dollar-control (arg &optional beg end cmd)
  "Insert GAMS dollar control option with completion.
List of candidates is created from elements of `gams-dollar-control-up'
and `gams-user-dollar-control-list' (and `gams-statement-mpsge'
if `gams-use-mpsge' is non-nil)."
  
  (interactive "P")
  (gams-statement-update)
  ;; Need to modify this.
  (setq gams-flag-dollar nil)
  (unwind-protect
      (let*
	  ((gams-alist gams-dollar-control-alist)
	   (completion-ignore-case t)
	   key1
	   (source-window (selected-window))
	   guess
	   (statement
	    (or cmd
		(progn

		  (setq guess
			(if gams-dollar-control-upcase
			    (upcase gams-dollar-control-name)
			  (downcase gams-dollar-control-name)))

		  (gams-read-dollar-control
		   (if gams-simple-messages
		       (format "Dollar control (default $%s): " guess)
		     (if (> (minibuffer-depth) 0)
			 (format "(default %s)%s: "
				 guess
				 (format "[level:%d]" (minibuffer-depth)))
		       (format "Insert dollar control ($ or @ = $, default = $%s): $" guess)))
		   nil))))
	   (statement (if (string= statement "") guess statement))
	   (statement (if gams-flag-dollar "" statement))
	   );;let

	;;
	(if (null (equal statement ""))
	    (setq gams-dollar-control-name statement))
	;;
	(if gams-dollar-control-upcase
	    (setq statement (upcase statement))
	  (setq statement (downcase statement)))
	;; Register or not?
	(if (null (or (member (list statement) gams-dollar-control-alist)
		      (equal statement "")))
	    (progn
	      (message "Store `%s' for future use?  Type `y' if yes: " statement)
	      (setq key1 (read-char))
	      (if (equal key1 ?y)
		  (progn (setq statement (upcase statement))
			 (gams-register statement t))
		nil))
	  nil)
	;; Insert.
	(if gams-dollar-control-upcase
	    (setq statement (upcase statement))
	  (setq statement (downcase statement)))
	(insert-string (concat "$" statement)))
    (if (<= (minibuffer-depth) 0) (use-global-map global-map))
    (insert "")))	;insert dummy string to fontify(Emacs20)

(defun gams-regexp-update ()
  "Update font-lock regular expression for GAMS statement
from `gams-user-statement-list' and `gams-statement-up'." 
  (setq gams-statement-regexp
	(concat
	 "^[ \n\t]*\\(zzz"
 	 (gams-create-regexp gams-user-statement-list)
 	 (gams-create-regexp
 	  (gams-list-downcase gams-user-statement-list))
	 (gams-create-regexp gams-statement-regexp-default)
	 (gams-create-regexp
	  (gams-list-downcase gams-statement-regexp-default))
	 "\\)\\([ \t\n$(]+\\)"
	 )))

;;; This function is from `font-latex.el'.
(defun gams-statement-regexp-func (limit)
  "Font lock regular expression for GAMS statement"
  (gams-regexp-update)
  (when (re-search-forward
	 gams-statement-regexp limit t)
    (let ((beg (match-beginning 1))
	  (end (match-end 1)))
      (store-match-data (list beg end))
      t)))

(defun gams-get-lst-filename ()
"Return the LST file name corresponding to the current GMS file buffer."
  (let ((file-buffer-gms (buffer-file-name))
	(ext-up (concat "." (upcase gams-lst-extention)))
	(ext-down (concat "." (downcase gams-lst-extention)))
	dir-gms	file-noext file-lst file-gms)
    ;; Store the GMS file name.
    (setq dir-gms (file-name-directory file-buffer-gms))
    (setq file-gms (file-name-nondirectory file-buffer-gms))
    (setq file-noext (file-name-sans-extension file-gms))
    ;; Search the LST file name
    (cond
     ((file-exists-p
       (concat dir-gms (upcase file-noext) ext-up))
      (setq file-lst (concat dir-gms (upcase file-noext) ext-up)))
     ((file-exists-p
       (concat dir-gms file-noext ext-down))
      (setq file-lst (concat dir-gms file-noext ext-down)))
     ((file-exists-p
       (concat dir-gms file-noext ext-up))
      (setq file-lst (concat dir-gms file-noext ext-up)))
     ((file-exists-p
       (concat dir-gms (upcase file-noext) ext-down))
      (setq file-lst (concat dir-gms (upcase file-noext) ext-down)))
     ((file-exists-p
       (concat dir-gms (downcase file-noext) ext-down))
      (setq file-lst (concat dir-gms (downcase file-noext) ext-down)))
     ((file-exists-p
       (concat dir-gms (downcase file-noext) ext-up))
      (setq file-lst (concat dir-gms (downcase file-noext) ext-up)))
     (t
      (message "LST file does not exist!")))
    ;; Return the name.
    file-lst))

(defun gams-view-lst ()
  "Switch to the LST file buffer and show the error message if any."
  (interactive)
  (let ((file-lst (gams-get-lst-filename)))
    ;;
    (if file-lst
	;; If the LST file exists.
	(progn
	 (let ((lst-buffer))
	   (if (get-file-buffer file-lst)
	       ;; If file-lst is already opened.
	       (progn
		 (set-buffer (get-file-buffer file-lst))
		 (if (verify-visited-file-modtime (current-buffer))
		     ;; If lst file is not changed
		     (progn
		       (switch-to-buffer (current-buffer))
		       ;; View error.
		       (gams-lst-view-error))
		   
		   ;; If lst file is chenged, kill-buffer.
		   (set-buffer-modified-p nil)
		   (kill-buffer (get-file-buffer file-lst))
		   (find-file file-lst)
		   (goto-char (point-min))
		   (gams-lst-mode)
		   (gams-lst-view-error)))
	     ;; if file-lst isn't opened.
	     (find-file file-lst)
	     (goto-char (point-min))
	     (gams-lst-mode)
	     (gams-lst-view-error))))
      ;; If the LST file not exits. 
      (message "The LST file does not exist!"))))


(defun gams-jump-to-lst ()
  "Switch to the LST file buffer."
  (interactive)
  (let ((file-lst (gams-get-lst-filename)))
    (if file-lst
	;; If lst file exists
	(progn
	  (let ((lst-buffer))
	    ;; lst file is already opened?
	    (if (get-file-buffer file-lst)
		;; If file-lst is already opened.

		;; lst file is modified?
		(if (verify-visited-file-modtime
		     (get-file-buffer file-lst))
		    ;; If not modified.
		    (pop-to-buffer (get-file-buffer file-lst))
		  ;; If modified.
		  (set-buffer-modified-p nil)
		  (kill-buffer (get-file-buffer file-lst))
		  (find-file find-file)
		  (gams-lst-mode))
	      ;; If file-lst isn't opened, open it.
	      (find-file file-lst)
	      (gams-lst-mode)))
	  (recenter))
      ;; LST file does not exits.
      (message "The LST file does not exist!"))))

;;; Comment insertion.
(defun gams-insert-comment ()
  "Insert a comment template defined by `gams-user-comment'."
  (interactive)
  (let ((point-a (point))
	(use-comment gams-user-comment)
	point-b point-c)
    (insert-string gams-user-comment)
    (setq point-b (point))
    (goto-char point-a)
    (if (re-search-forward "%" point-b t)
	(progn (setq point-c (match-beginning 0))
	       (replace-match ""))
      (goto-char point-b))))

;;; From yatex.el
(defun gams-insert-parens (arg)
  "Insert parenthesis pair."
  (interactive "p")
  (if gams-close-paren-always
      (progn (insert "()")
	     (backward-char 1))
    (insert "(")))

(defun gams-insert-double-quotation (arg)
  "Insert double quotation pair."
  (interactive "p")
  (if gams-close-double-quotation-always
      (progn (insert "\"\"")
	     (backward-char 1))
    (insert "\"")))

(defun gams-insert-single-quotation (arg)
  "Insert single quotation pair."
  (interactive "p")
  (if gams-close-single-quotation-always
      (progn (insert "`'")
	     (backward-char 1))
    (insert "`")))

;;; Fill paragraph function.  This is from "lisp-mode.el"
;;; (`lisp-fill-paragraph').  I changed ";" in the original function to
;;; "\\(*\\)".  This function is likely not to work well in many cases.

(defun gams-fill-paragraph (&optional justify)
  "Like \\[fill-paragraph], but handle GAMS comment.
If any of the current line is a comment, fill the comment or the
paragraph of it that point is in, preserving the comment's indentation
and initial *."
  (interactive "P")
  (let (
	;; Non-nil if the current line contains a comment.
	has-comment
	;; Non-nil if the current line contains code and a comment.
	has-code-and-comment
	;; If has-comment, the appropriate fill-prefix for the comment.
	comment-fill-prefix
	)

    ;; Figure out what kind of comment we are looking at.
    (save-excursion
      (beginning-of-line)
      (cond
       ;; A line with nothing but a comment on it?
       ((looking-at "^\\(*\\)[* \t]*")
	(setq has-comment t
	      comment-fill-prefix (buffer-substring (match-beginning 0)
						    (match-end 0))))
       ))
    (if (not has-comment)
        ;; `paragraph-start' is set here (not in the buffer-local
        ;; variable so that `forward-paragraph' et al work as
        ;; expected) so that filling (doc) strings works sensibly.
        ;; Adding the opening paren to avoid the following sexp being
        ;; filled means that sexps generally aren't filled as normal
        ;; text, which is probably sensible.  The `;' and `:' stop the
        ;; filled para at following comment lines and keywords
        ;; (typically in `defcustom').
	(let ((paragraph-start (concat paragraph-start "")))
          (fill-paragraph justify))
      ;; Narrow to include only the comment, and then fill the region.
      (save-excursion
	(save-restriction
	  (beginning-of-line)
	  (narrow-to-region
	   ;; Find the first line we should include in the region to fill.
	   (save-excursion
	     (while (and (zerop (forward-line -1))
			 (looking-at "^\\(*\\)")))
	     ;; We may have gone too far.  Go forward again.
	     (or (looking-at "^\\(*\\)")
		 (forward-line 1))
	     (point))
	   ;; Find the beginning of the first line past the region to fill.
	   (save-excursion
	     (while (progn (forward-line 1)
			   (looking-at "^\\(*\\)")))
	     (point)))
	  ;; Lines with only * on them can be paragraph boundaries.
	  (let* ((paragraph-start (concat paragraph-start "\\|^\\(*\\)$"))
		 (paragraph-separate (concat paragraph-start "\\|^\\(*\\)$"))
		 (paragraph-ignore-fill-prefix nil)
		 (fill-prefix comment-fill-prefix)
		 (after-line (if has-code-and-comment
				 (save-excursion
				   (forward-line 1) (point))))
		 (end (progn
			(forward-paragraph)
			(or (bolp) (newline 1))
			(point)))
		 ;; If this comment starts on a line with code,
		 ;; include that like in the filling.
		 (beg (progn (backward-paragraph)
			     (if (eq (point) after-line)
				 (forward-line -1))
			     (point))))
	    (fill-region-as-paragraph beg end
				      justify nil
				      (save-excursion
					(goto-char beg)
					(if (looking-at fill-prefix)
					    nil
					  (re-search-forward comment-start-skip)
					  (point))))))))
    t))


;;; Process handling.

;;; Most of the codes for process handling are from epo.el ,epolib.el,
;;; epop.el in the `EPO' package written by Yuuji Hirose.  I modified
;;; them.

;;; From epolib.el
(defun gams*window-list ()
  (let*((curw (selected-window)) (win curw) (wlist (list curw)))
    (while (not (eq curw (setq win (next-window win))))
      (or (eq win (minibuffer-window))
	  (setq wlist (cons win wlist))))
    wlist))

(defun gams*smart-split-window (height)
  "Split current window wight specified HEIGHT.
If HEIGHT is number, make a new window that has HEIGHT lines.
If HEIGHT is string, make a new window that occupies HEIGT % of screen height.
Otherwise split window conventionally."
  (if (one-window-p t)
      (split-window
       (selected-window)
       (max
        (min
         (- (gams*screen-height)
            (if (numberp height)
                (+ height 2)
              (/ (* (gams*screen-height)
                    (string-to-int height))
                 100)))
         (- (gams*screen-height) window-min-height 1))
        window-min-height))))

(defun gams*showup-buffer (buffer &optional select)
  "Make BUFFER show up in certain window (except selected window).
Non-nil for optional argument SELECT keeps selection to the target window."
  (let (w)
    (or
     ;;if already visible
     (if gams:frame-feature-p
	 (if (setq w (get-buffer-window buffer t))
	     (if select
		 (progn
		   (raise-frame (select-frame (window-frame w)))
		   (set-mouse-position (selected-frame) 0 -1))
	       w))
       ;;no frames
       (if (setq w (get-buffer-window buffer))
	   (if select (select-window w) w)))
     ;;not visible
     (let ((sw (selected-window))
	   (wlist (gams*window-list)))
       (cond
	((eq (current-buffer) (get-buffer buffer)) nil)
	((one-window-p)
	 (gams*smart-split-window gams-default-pop-window-height)
	 (select-window (next-window nil 1))
	 (switch-to-buffer (get-buffer-create buffer)))
	((= (length wlist) 2)
	 (select-window (get-lru-window))
	 (switch-to-buffer (get-buffer-create buffer)))
	(t   ;more than 2windows
	 (select-window (next-window nil 1))
	 (switch-to-buffer (get-buffer-create buffer))))
       (or select (select-window sw))))))


;;; From epop.el
(defun gams*process-sentinel (proc mess)
  "Display the end of process buffer."
  (cond
   ((memq (process-status proc) '(signal exit))
    (save-excursion
      (let ((sw (selected-window)) w)
	(set-buffer (process-buffer proc))
	(goto-char (point-max))
	(insert
	 (format "\nGAMS process finished at %s\n" (current-time-string)))
	(cond
	 ((and gams:frame-feature-p
	       (setq w (get-buffer-window (current-buffer) t)))
	  (select-frame (window-frame w))
	  (select-window w)
	  (goto-char (point-max))
	  (recenter -1))
	 ((setq w (get-buffer-window (current-buffer)))
	  (select-window w)
	  (goto-char (point-max))
	  (recenter -1)))
	(select-window sw))
      (message "GAMS process has done.  Type `C-cv' if you view a LST file.")))))

;;

(defun gams*start-process-other-window (name commandline)
  "Start command line (via shell) in the next window."
  (let (
;	(b (get-buffer-create "*Running GAMS*"))
	(sw (selected-window)) p
	(dir default-directory)
	pbuff-name)
;    (set (make-local-variable 'gams:current-process-buffer) b)
    (setq pbuff-name (gams-get-process-buffer))
      
    (gams*showup-buffer pbuff-name t) ;popup buffer and select it.
    (message "Starting GAMS process...")
    (kill-all-local-variables)
    (erase-buffer)
    (cd dir)
    (setq default-directory dir)
    (insert commandline "\n ")
    (set (make-local-variable 'gams:process-command-name) name)
    (set-process-sentinel
     (setq p (start-process
	      name
	      pbuff-name
	      shell-file-name
	      gams:shell-c commandline))
     'gams*process-sentinel)
    (set-marker (process-mark p) (1- (point)))
    (setq major-mode 'gams-process-mode)
;    (use-local-map gams-process-mode-map)
    (select-window sw)))

(defun gams*get-builtin (keyword)
  "Get built-in string specified by KEYWORD in current buffer."
  (save-excursion
    (save-restriction
      (widen)
      (goto-char (point-min))
      (if (search-forward
	   (concat
	    comment-start	   ;buffer local variable in any buffer
	    keyword) nil t)
	  (let ((peol (progn (end-of-line) (point))))
	    (gams*buffer-substring
	     (progn
	       (goto-char (match-end 0))
	       (skip-chars-forward " \t")
	       (point))
	     (if (and comment-end
		      (stringp comment-end)
		      (string< "" comment-end)
		      (re-search-forward
		       (concat (regexp-quote comment-end)
			       "\\|$")
		       peol 1))
		 (match-beginning 0)
	       peol)))))))

(defun gams*update-builtin (keyword newdef)
  "Update built-in KEYWORD to NEWDEF"
  (save-excursion
    (save-restriction
      (widen)
      (goto-char (point-min))
      (if (search-forward (concat comment-start keyword) nil t)
	  (let ((peol (progn (end-of-line) (point))))
	    (goto-char (match-end 0))
	    (skip-chars-forward " \t")
	    (delete-region
	     (point)
	     (if (and comment-end (stringp comment-end)
		      (string< "" comment-end)
		      (search-forward comment-end peol t))
		 (progn (goto-char (match-beginning 0)) (point))
	       peol))
	    (insert newdef))
	(while (and (progn (skip-chars-forward " \t")
			   (looking-at (regexp-quote comment-start)))
		    (not (eobp)))
	  (forward-line 1))
	(open-line 1)
	(insert comment-start keyword newdef comment-end)))))

(defun gams-get-program-filename (&optional nodir noext)
  ""
  (cond
   ((and nodir (null noext)
    (file-name-nondirectory (buffer-file-name (current-buffer)))))
   ((and (null nodir) noext)
    (file-name-sans-extension (buffer-file-name (current-buffer))))
   ((and nodir noext)
    (file-name-nondirectory
     (file-name-sans-extension (buffer-file-name (current-buffer)))))
   (t
    (buffer-file-name (current-buffer)))
    ))

(defun gams-get-process-buffer ()
  ""
  (let ((flag-a gams-multi-process)
	pro-buf-name)
    (if flag-a
	;; Multi-process.
	(get-buffer-create
	 (concat gams*command-process-buffer " on "
		 (gams-get-program-filename t) "*"))
      ;; Not multi-process.
      (get-buffer-create (concat gams*command-process-buffer "*"))
      )))

(defun gams*start-processor (&optional ask)
  "Start GAMS on the current file."
  (let* (prompt
	(builtin "#!")
	(command "compile")
	arg string newarg)
    (setq arg
	  (or
	   ;; if built-in processor specified, use it
	   (and builtin (gams*get-builtin builtin))
	   (concat gams:process-command-name " "
		   (file-name-nondirectory buffer-file-name) " "
		   gams:process-command-option)))
    (basic-save-buffer)
    ;(setq arg (concat command " " arg))
    (gams*start-process-other-window
     command
     (cond
      (prompt
       (read-string "Execute: " arg))
      (ask
       (setq newarg (read-string "Edit command if you want:  " arg))
       (if (and builtin
		(not (string= newarg arg))
		(y-or-n-p "Use this command line also in the future? "))
	   (progn
	     (gams*update-builtin builtin newarg)
	     (message "The command line is inserted in the fisrt line in this file!")))
	 newarg)
      (t arg)))))

(defun gams*kill-processor ()
  "Stop (kill) a GAMS process."
  (let ((p (get-buffer-process (gams-get-process-buffer))))
    (if p
	(interrupt-process p)
      (message "GAMS process has already exited."))))


(defun gams-start-menu (&optional ask char)
  "Process starting menu.
Optional second argument CHAR is for non-interactive call from menu."
  (interactive "P")
  (message (format "Start GAMS (%c) or Kill GAMS process (%c): "
		   gams-run-key gams-kill-key))
  (let ((c (or char (read-char))))
    ;;
    (cond
     ((equal c gams-run-key)
      (gams*start-processor ask))
     ((equal c gams-kill-key)
      (gams*kill-processor))
     (t
      (message "No such choice `%c'" c)))))

(defun gams-recenter ()
  "Recentering."
  (interactive)
  (if gams-recenter-font-lock
      (cond
       (gams-emacs-21
	(font-lock-mode -1)
	(font-lock-mode 1))
       (gams-emacs-20
	(font-lock-mode -1)
	(font-lock-mode 1))
       ))
  (recenter))

;;; Jump between ontext and offtext.
(defun gams-judge-on-off-text ()
  "Judge whether curson is on $ontext or $offtext.

$ontext => return on,
$offtext => return off,
Otherwise => return nil."
  (let (point-beg
	temp-text
	temp-flag)
    (save-excursion
      (if (not (equal (char-to-string (char-after)) "$"))
	  (re-search-backward "\\([$]\\|^[ \t]\\|[ \n]\\)" nil t))
      (setq point-beg (+ 1 (point)))
      (forward-word 1)
      (setq temp-text
	    (downcase (buffer-substring point-beg (point))))
      (cond
       ((equal temp-text "ontext")
	(setq temp-flag "on"))
       ((equal temp-text "offtext")
	(setq temp-flag "off"))
       (t
	(setq temp-flag nil)))
      temp-flag)))

(defun gams-jump-on-off-text ()
  "Jump between $ontext-$offtext.

If you execute this command on $ontext ($offtext), then you jump to the
corresponding $offtext ($ontext).
"
  (interactive)
  (let ((flag-on (gams-judge-on-off-text))
	(init-point (point))
	(case-fold-search t)
	temp-a)
  ;; ontext or offtext.
  (cond
   ;; ontext.
   ((equal flag-on "on")
    (forward-word 1)
    (if (re-search-forward "$\\(ontext\\|offtext\\)" nil t)
	(progn
	  (setq temp-a (downcase (buffer-substring (match-beginning 1)
						   (match-end 1))))
	  (cond
	   ((equal temp-a "ontext")
	    (goto-char init-point)
	    (message "No corresponding $offtext exists!"))
	   ((equal temp-a "offtext")
	    (forward-word -1)
	    (forward-char 1)
	    (message "The corresponding $offtext is found!"))))
      (message "No corresponding $offtext exists!")))
   ;; offtext.
   ((equal flag-on "off")
    (forward-word -1)
    (if (re-search-backward "$\\(ontext\\|offtext\\)" nil t)
	(progn
	  (setq temp-a (downcase (buffer-substring (match-beginning 1)
					 (match-end 1))))
	  (cond
	   ((equal temp-a "ontext")
	    (forward-char 1)
	    (message "The corresponding $ontext is found!"))
	   ((equal temp-a "offtext")
	    (goto-char init-point)
	    (message "No corresponding $ontext exits!"))))
      (message "No corresponding $ontext exits!")))
   ;; Otherwise.  Do nothing.
   (t
    (message "This command is valid only if the cursor is on either $ontext or $offtext."))
    )))


;;; View manuals.

(defvar gams-read-docs-history nil "Holds history of dollar control.")
(put 'gams-read-dollar-history 'no-default t)

(defvar gams-read-doc-completion-map nil
  "*Key map for gams-read-docs.")
(if gams-read-doc-completion-map nil
   (setq gams-read-doc-completion-map
	 (copy-keymap minibuffer-local-completion-map))
   (define-key gams-read-doc-completion-map
     " " 'gams-minibuffer-complete)
   (define-key gams-read-doc-completion-map
     "\C-i" 'gams-minibuffer-complete))

(defun gams-read-docs (prompt &optional predicate initial)
  "Read a GAMS dollar control operation with completion."
  (let ((minibuffer-completion-table
	 (append gams-manuals-alist)))
    (read-from-minibuffer
     prompt initial gams-read-doc-completion-map nil
     'gams-read-docs-history)))

(defun gams-view-docs ()
  "View GAMS manuals.

Envoke the pdf file viewer and see GAMS manuals."
  (interactive)
  (unwind-protect
      (let*
	  ((gams-alist gams-manuals-alist)
	   (completion-ignore-case t)
	   key1
	   (source-window (selected-window))
	   guess
	   (statement
	    (progn
	      (setq guess "User-Manual")
	      (gams-read-docs
	       (format "View which manual? (default = %s): " guess))))
	   (statement (if (string= statement "") guess statement))
	   (statement (if gams-flag-dollar "" statement))
	   ;; directory name.
	   (dir-name
	    (if (string-match "/" gams-docs-dir (string-width gams-docs-dir))
		(substring gams-docs-dir 1 (- (string-width gams-docs-dir)))
	      gams-docs-dir))
	   file-name
	   dir-docs
	   (prog-name gams-docs-view-program)
	   (buf (get-buffer-create "*View GAMS manual*"))
	   proc
	   );;let* ends.
	(if (string-match "Solver" statement)
	    (setq dir-docs "/solvers/")
	  (setq dir-docs "/gams/"))
	(setq file-name (concat dir-name dir-docs
				(cdr (assoc statement gams-manuals-alist))
				".pdf"))
	;; Start process.
	(setq proc (start-process "manual" buf prog-name file-name " &"))
	(message "Starting the pdf file viewer...")
	)))


;;;	From yatex.el

;;; autoload
(defun substitute-all-key-definition (olddef newdef keymap)
  "Replace recursively OLDDEF with NEWDEF for any keys in KEYMAP now
defined as OLDDEF. In other words, OLDDEF is replaced with NEWDEF
where ever it appears."
    (mapcar
     (function (lambda (key) (define-key keymap key newdef)))
     (where-is-internal olddef keymap)))
;;-------------------- Final hook jobs --------------------
(substitute-all-key-definition
 'fill-paragraph 'gams-fill-paragraph gams-mode-map)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;	Code for GAMS-LST mode.
;;;	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; key assignment.
(defvar gams-lst-mode-map (make-keymap) "Keymap for gams-lst-mode")
(let ((map gams-lst-mode-map))
  (define-key map gams-lk-3 'gams-lst-view-error)
  (define-key map gams-lk-2 'gams-lst-jump-to-error-file)  
  (define-key map gams-lk-1 'gams-lst-jump-to-input-file)
  (define-key map "q" 'gams-lst-kill-buffer)    
  (define-key map "?" 'gams-lst-help)    

  (define-key map "o" 'gams-outline)

  (define-key map "s" 'gams-lst-solve-summary)
  (define-key map "S" 'gams-lst-solve-summary-back)
  (define-key map "r" 'gams-lst-report-summary)    
  (define-key map "R" 'gams-lst-report-summary-back)    
  (define-key map "v" 'gams-lst-next-var)    
  (define-key map "V" 'gams-lst-previous-var)    
  (define-key map "e" 'gams-lst-next-equ)    
  (define-key map "E" 'gams-lst-previous-equ)    
  (define-key map "p" 'gams-lst-next-par)    
  (define-key map "P" 'gams-lst-previous-par)    

  (define-key map "l" 'gams-lst-query-jump-to-line)    
  (define-key map gams-lk-5 'gams-lst-jump-to-line)    

  (define-key map " " 'scroll-up)    
  (define-key map [delete] 'scroll-down)    
  (define-key map "1" 'gams-lst-widen-window)    
  (define-key map "2" 'gams-lst-split-window)
  (define-key map "m" 'gams-lst-move-frame)    
  (define-key map "w" 'gams-lst-resize-frame)
  (define-key map "z" 'gams-lst-move-cursor)    

  (define-key map gams-lk-4 'gams-lst-jump-to-input-file-2)

  (define-key map "d" 'gams-lst-scroll-1)
  (define-key map "f" 'gams-lst-scroll-down-1)
  (define-key map "g" 'gams-lst-scroll-2)
  (define-key map "h" 'gams-lst-scroll-down-2)
  (define-key map "j" 'gams-lst-scroll-double)
  (define-key map "k" 'gams-lst-scroll-down-double)

  (define-key map "D" 'gams-lst-scroll-page-1)
  (define-key map "F" 'gams-lst-scroll-page-down-1)
  (define-key map "G" 'gams-lst-scroll-page-2)
  (define-key map "H" 'gams-lst-scroll-page-down-2)
  (define-key map "J" 'gams-lst-scroll-page-double)
  (define-key map "K" 'gams-lst-scroll-page-down-double)
  )

;;; Menu for GAMS-LST mode.
(easy-menu-define 
  gams-lst-menu gams-lst-mode-map "Menu keymap for GAMS-LST mode."
  '("GAMS-LST"
    ["Jump to the error and show its meaning" gams-lst-view-error t]
    ["Jump to the error place in the program file" gams-lst-jump-to-error-file t]
    ["Jump to the input file" gams-lst-jump-to-input-file t]
    ["Close the buffer" gams-lst-kill-buffer t]
    ["Show help" gams-lst-help t]
    "--"
    ["Start the GAMS-OUTLINE mode" gams-outline t]
    "--"
    ["Jump to a line you specify" gams-lst-query-jump-to-line t]    
    ["Jump to a line" gams-lst-jump-to-line t]
    "--"
    ["Jump to the next SOLVER SUMMARY" gams-lst-solve-summary t]    
    ["Jump to the next REPORT SUMMARY" gams-lst-report-summary t]    
    ["Jump to the next VAR entry" gams-lst-next-var t]    
    ["Jump to the next EQU entry" gams-lst-next-equ t]    
    ["Jump to the next PARAMETER entry" gams-lst-next-par t]    
    ))


(defun gams-lst-mode ()
  "Major mode for viewing GAMS LST file. 

The following commands are available in the GAMS-LST mode:

\\[gams-lst-view-error]		Jump to the error and show its number and meaning.
\\[gams-lst-jump-to-error-file]		Jump back to the error place in the program file.
\\[gams-lst-jump-to-input-file] 		Jump to the input (GMS) file.
\\[gams-lst-kill-buffer]		Close the buffer.
\\[gams-lst-help]		Display this help.

\\[gams-outline]		Start the GAMS-OUTLINE mode.

\\[gams-lst-solve-summary](\\[gams-lst-solve-summary-back])		Jump to the next (previous) SOLVER SUMMARY.
\\[gams-lst-report-summary](\\[gams-lst-report-summary-back])		Jump to the next (previous) REPORT SUMMARY.
\\[gams-lst-next-var](\\[gams-lst-previous-var])		Jump to the next (previous) VAR entry.
\\[gams-lst-next-equ](\\[gams-lst-previous-equ])		Jump to the next (previous) EQU entry.
\\[gams-lst-next-par](\\[gams-lst-previous-par])		Jump to the next (previous) PARAMETER entry.

\\[gams-lst-query-jump-to-line]		Jump to a line you specify.
\\[gams-lst-jump-to-line]		Jump to a line.

\\[scroll-up]		Scroll up.
\\[scroll-down] or DEL	Scroll down.
\\[gams-lst-widen-window]		Widen the window.
\\[gams-lst-split-window]		Split the window.
\\[gams-lst-move-frame]		Move frame.
\\[gams-lst-resize-frame]		Resize frame.
\\[gams-lst-move-cursor]		Move a cursor to the other window.


[Commands for Scrolling.]

Suppose that there are two windows displayed like

    __________________	  
   |   	      	      |  
   |  LST buffer 1    |  ==>  LST-1.
   |   	      	      |  
   |  CURSOR  here    |  
   |   	      	      |  
   |------------------|  
   |		      | 
   |  LST buffer 2    |  ==>  LST-2.
   |   	      	      |  
   |		      | 
    ------------------

'\\[gams-lst-scroll-1](\\[gams-lst-scroll-down-1])		Scroll the current buffer (LST-1) up (down) one line.
'\\[gams-lst-scroll-2](\\[gams-lst-scroll-down-2])		Scroll the next buffer (LST-2) up (down) one line.
'\\[gams-lst-scroll-double](\\[gams-lst-scroll-down-double])		Scroll two buffers (LST-1 and LST-2) up (down) one line.

Keyboard.

  _____________________________________________________________
  |         |         |         |         |         |         |
  |    d    |    f    |    g    |    h    |    j    |    k    |
  |         |         |         |         |         |         |
  -------------------------------------------------------------

       |         |         |         |         |         |

      UP        DOWN      UP        DOWN      UP        DOWN
         LST-1               LST-2             LST-1 & 2

If only one window exists, the above three commands have the same function
(scroll up/down the current buffer).

The followings are page scroll commands.  Just changed to upper cases.

'\\[gams-lst-scroll-page-1](\\[gams-lst-scroll-page-down-1])		Scroll up (down) the current buffer (LST-1) by a page.
'\\[gams-lst-scroll-page-2](\\[gams-lst-scroll-page-down-2])		Scroll up (down) the next buffer (LST-2) by a page.
'\\[gams-lst-scroll-page-double](\\[gams-lst-scroll-page-down-double])		Scroll up (down) two buffers (LST-1 and LST-2) by a page.
"   		       	
  (interactive)
  (setq major-mode 'gams-lst-mode)
  (setq mode-name "GAMS-LST")
  (use-local-map gams-lst-mode-map)
;  (easy-menu-add todo-menu)
;  (outline-minor-mode 1)
;  (hide-other)
  (setq buffer-read-only t) ;make the buffer read-only.
  
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(gams-lst-font-lock-keywords t t))
  (font-lock-fontify-buffer)

;; Create several buffer local variables for the OUTLINE mode.

  (make-local-variable 'gams-ol-flag)
  (setq gams-ol-flag nil)

  (make-local-variable 'gams-over-alist)
  (setq gams-over-alist nil)

  (make-local-variable 'gams-ol-alist-tempo)
  (setq gams-ol-alist-tempo nil)

  ;;
  (easy-menu-add gams-lst-menu)

  (run-hooks 'gams-lst-mode-hook))
;; gams-lst-mode ends here.


(defun gams-lst-help ()
  "Display help for the GAMS-LST mode."
  (interactive)
  (describe-function 'gams-lst-mode))

(defun gams-lst-kill-buffer ()
  "Close the LST buffer"
  (interactive)
  (let ((ov-buff (concat "*" (file-name-sans-extension
			      (file-name-nondirectory
			       (buffer-file-name))) "-OUTLINE*")))
    (if (get-buffer ov-buff)
	(kill-buffer ov-buff))
  (kill-buffer nil)))

(defun gams-lst-view-error ()
  "Move to the error place.
and show its meaning in another window if error number is displayed."
  (interactive)
  (goto-char (point-min))
  (let (error-num error-place error-mes-place error-column)
    ;; First search syntax error. 
    (if (re-search-forward "\\*\\*\\*\\* [ ]+\\(\\$\\)\\([0-9]+\\)[$]?" nil t)
	(progn
	  (goto-char (match-beginning 1))
	  (setq error-place (point))
	  ;; set `error-num' the found error number. It is nil if no error.
	  (setq error-num (buffer-substring (match-beginning 2)
					    (match-end 2)))
	  (message "Key: `%s' = Jump to the error place, `%s' = Jump to the input file."
		   gams-lk-2 gams-lk-1)
	  ;;
	  (if error-num
	      (progn
		(if (null (re-search-forward "Error Messages" nil t))
		    (message "cannot find `Error Messages'!")
		  (setq error-mes-place
			(re-search-forward error-num nil t))))
	    ;; if error-num is nil, go to the top of the buffer.
	    (goto-char (point-min)))
	  ;; Display syntax error message.
	  (if error-mes-place
	      (progn
		(delete-other-windows)
		(split-window)
		(goto-char error-place)
		(recenter)
		(other-window 1)
		(goto-char error-mes-place)
		(recenter 0)
		(other-window 1))))

      ;; Search another type of errors.
      (if (catch 'found
	    (while (re-search-forward "\\*\\*\\*\\* " nil t)
	      (progn
		(beginning-of-line)
		(setq a-point (point))
		(end-of-line)
		(setq b-point (point))
		(goto-char a-point)

		;; The following lines are not regarded as errors and
		;; skipped.  Is this right behavior?
		(if (not (or (re-search-forward "\\*\\*\\*\\* SOLVER STATUS"
					    b-point t)
			     (re-search-forward "\\*\\*\\*\\* MODEL STATUS"
					    b-point t)
			     (re-search-forward "\\*\\*\\*\\* REPORT SUMMARY"
					    b-point t)
			     (re-search-forward "\\*\\*\\*\\* REPORT FILE SUMMARY"
					    b-point t)
			     (re-search-forward "\\*\\*\\*\\* LIST OF STRAY NAMES"
					    b-point t)
			     (re-search-forward "\\*\\*\\*\\* STRAY NAME "
					    b-point t)
			     (re-search-forward "\\*\\*\\*\\* FILE SUMMARY"
					    b-point t)
			     (re-search-forward "\\*\\*\\*\\* OBJECTIVE VALUE"
					    b-point t)))
		    (throw 'found t)
		  (forward-line 1)))))
	  ;;
	  (progn
	    (goto-char (match-beginning 0))
	    (setq a-point (point))
	    (end-of-line)
	    (setq b-point (point))
	    (goto-char a-point)
	    ;;
	    (if (re-search-forward " at line \\([0-9]+\\)" b-point t)
		(message (format "Error is found!  Type `%s' if you want to jumpt to the error line %s."
				 gams-lk-5 (buffer-substring (match-beginning 1)
							     (match-end 1))))
	    (message "Error is found!"))
	    (goto-char a-point))

	;; Else part.  When no error is found.
	(progn
	  (message "No error message is found!")
	  (goto-char (point-min)))
	))))

;;; 
(defun gams-lst-jump-to-error ()
  "Jump to the error place."
  (interactive)
  (let ((current-point (point)))
    (goto-char (point-min))
    (if	(re-search-forward "\\*\\*\\*\\* [ ]+\\(\\$\\)\\([0-9]+\\)[$]?" nil t)
	(progn
	  (goto-char (match-beginning 1))
	  t)
      (progn
	(goto-char current-point)
	(message "No error is found!")
	nil))))

(defun gams-lst-get-gms ()
  "Return a GMS file name from a the current LST file buffer."
  (let ((file-buffer-lst (buffer-file-name))
	(ext-up (concat "." (upcase gams-lst-gms-extention)))
	(ext-down (concat "." (downcase gams-lst-gms-extention)))
	dir-lst file-noext file-gms file-lst)
    
    ; Store LST file name.
    (setq dir-lst (file-name-directory file-buffer-lst))
    (setq file-lst (file-name-nondirectory file-buffer-lst))
    (setq file-noext (file-name-sans-extension file-lst))

    ; Search GMS file name.  GMS file name is stored in file-gms.
    (cond
     ((file-exists-p
       (concat dir-lst file-noext ext-down))
      (setq file-gms (concat dir-lst file-noext ext-down)))
     ((file-exists-p
       (concat dir-lst file-noext ext-up))
      (setq file-gms (concat dir-lst file-noext ext-up)))
     ((file-exists-p
       (concat dir-lst (upcase file-noext) ext-down))
      (setq file-gms (concat dir-lst (upcase file-noext) ext-down)))
     ((file-exists-p
       (concat dir-lst (upcase file-noext) ext-up))
      (setq file-gms (concat dir-lst (upcase file-noext) ext-up)))
     ((file-exists-p
       (concat dir-lst (downcase file-noext) ext-down))
      (setq file-gms (concat dir-lst (downcase file-noext) ext-down)))
     ((file-exists-p
       (concat dir-lst (downcase file-noext) ext-up))
      (setq file-gms (concat dir-lst (downcase file-noext) ext-up)))
     (t
      (message "GMS file does not exist!")))
    file-gms))

(defun gams-lst-get-input-filename ()
  "Get the input file name associated to the current LST file.

The input file name is extract from FILE SUMMARY field."
  (let (temp-file)
    (save-excursion
      (goto-char (point-min))
      (if (re-search-forward "\\*\\*\\*\\* FILE SUMMARY" nil t)
	  ;; If FILE SUMMARY is found,
	  (progn
	    (setq point-a (match-beginning 0))
	    
	    (if (re-search-forward "^INPUT[ ]+\\(.*\\)" nil t)
		(setq temp-file (buffer-substring (match-beginning 1)
						  (match-end 1)))))
	;; If FILE SUMMARY is not found,
	(setq temp-file (gams-lst-get-gms))
	(message "FILE SUMMARY field does not exits!  The extension is assumed to be gms.")
	(sleep-for 1)))
    ;; Return the input file name.
    temp-file))

(defun gams-lst-save-string ()
  "?"
  (let ((times 3)
	(check 0)
	(current-buffer (buffer-name))
	(work-buffer (get-buffer-create "*temp*")) ; Create temporary buffer
	point-beg point-end list-string string
	a-point b-point)
    ;; save string in list-string
    (while (and (> times 0) (< check 50))
      (forward-line -1)
      (beginning-of-line)
      (setq point-beg (point))
      (end-of-line)
      (setq point-end (point))
      (goto-char point-beg)
      (if (re-search-forward "^[ ]*[0-9]+  " point-end t)
	  (progn
	    (setq list-string
		  (cons
		   (if (equal
			(buffer-substring (match-end 0) point-end) " ")
		       ;; if t, return "" not " ".
		       ""
		     ;; if nil, return the matched.
		     (buffer-substring (match-end 0) point-end))
		   list-string))
	    (setq times (- times 1)))
	(setq check (+ check 1))
	))
;;
    (setq string
	  (concat
	   (car list-string) "\n"
	   (car (cdr list-string)) "\n"
	   (car (cdr (cdr list-string)))))

    ;; Switch to the temporary buffer.
    (set-buffer work-buffer)
    (insert string)
    (goto-char (point-min))

    ;; Replace several strings with regular expression.
    (goto-char (point-min))
    (while (re-search-forward "\\(\\\\\\)" nil t)
      (setq a-point (match-beginning 1))
      (setq b-point (match-end 1))
      (replace-match "[\\\\]"))
    (goto-char (point-min))
    (while (re-search-forward "\\(\\+\\)" nil t)
      (setq a-point (match-beginning 1))
      (setq b-point (match-end 1))
      (replace-match "[\+]"))
    (goto-char (point-min))
    (while (re-search-forward "\\(\\?\\)" nil t)
      (setq a-point (match-beginning 1))
      (setq b-point (match-end 1))
      (replace-match "[\?]"))
    (goto-char (point-min))
    (while (re-search-forward "\\([ ]+\\)" nil t)
      (setq a-point (match-beginning 1))
      (setq b-point (match-end 1))
      (replace-match "[ \t]+"))
    (goto-char (point-min))
    (while (re-search-forward "\\(\\.\\)" nil t)
      (setq a-point (match-beginning 1))
      (setq b-point (match-end 1))
      (replace-match "\."))
    (goto-char (point-min))     
    (while (re-search-forward "\\(\\$\\)" nil t)
      (setq a-point (match-beginning 1))
      (setq b-point (match-end 1))
      (replace-match "\$"))
    (goto-char (point-min))
    (while (re-search-forward "\\(\\*\\)" nil t)
      (setq a-point (match-beginning 1))
      (setq b-point (match-end 1))
      (replace-match "[\*]"))
    (goto-char (point-min))
    (while (re-search-forward "\\(\n\\)" nil t)
      (setq a-point (match-beginning 1))
      (setq b-point (match-end 1))
      (replace-match "[ \t]*\n"))
    ;; Store regular expression to string.
    (setq string (buffer-substring (point-min) (point-max)))
    (insert string)
    (switch-to-buffer current-buffer)
    ;; Delete temporary buffer.
    (kill-buffer work-buffer)
    ;; Return the content of string.
    string))

;;; This function is not used.
(defun gams-lst-view-gms()
  "Jump back to the error place in the program file.

NB: It may not work properly.  It is difficult to implement this function
perfectly...
"
  (interactive)
  (let ((file-gms (gams-lst-get-input-filename))
	string)
    
    ;; Jump to the error place.
    (if (gams-lst-jump-to-error)
	(progn
	  ;; Save the string around error place.
	  (setq string (gams-lst-save-string))
	  (if (null (file-exists-p file-gms))
	      ;;
	      (message "The file `%s' does not exist!" file-gms)
	    ;; open GMS file.
	    (if (get-file-buffer file-gms)
		(switch-to-buffer (get-file-buffer file-gms))
	      (find-file file-gms))
	    (goto-char (point-min))
	    ;; Search the error place.
	    (if (re-search-forward string nil t)
		(progn
		  (forward-line -1)
		  (recenter))
	      (message "Error place is not found!"))
	    )
	  (message "No error is found!")))))

;;; New function.
(defun gams-lst-jump-to-error-file ()
  (interactive)
  (let (point-a point-b line-num file-name error-column temp-col)
    ;;
    (if	(gams-lst-jump-to-error)
	(progn
	  ;; Store column number.
	  (save-excursion
	    (if (re-search-forward "^\\([ ]*[0-9]+[ ][ ]\\)" nil t)
		(setq temp-col (current-column))
	      (setq temp-col 1)))
	  (setq error-column (- (current-column) temp-col))
	  (forward-line 1)
	  (setq point-a (point))
	  (end-of-line)
	  (forward-char 1)
	  (setq point-b (point))
	  (goto-char point-a)
	  ;; Search the file name line.
	  (if (re-search-forward
	       "\\*\\*\\*\\* LINE[ \t]+\\([0-9]+\\)[ ]+IN[ ]+FILE[ ]+\\([^\n]+\\)\n"
	       point-b t)
	      ;; If the file name line is found.
	      (progn
		(setq line-num (buffer-substring (match-beginning 1)
						 (match-end 1)))
		(setq file-name (buffer-substring (match-beginning 2)
						  (match-end 2)))
		(if (file-exists-p file-name)
		    ;;
		    (progn
		      (if (get-file-buffer file-name)
			  (switch-to-buffer (get-file-buffer file-name))
			(find-file file-name))
		     
		      (goto-line (string-to-number line-num))
		      (move-to-column error-column)
		      (recenter))

		  ;;
		  (message "The file `%s' does not exist!" file-name)))
	  
	    ;; If the file name line is not found

	    ;; Jump to the error place.
	    (gams-lst-jump-to-error)
	    (recenter)

	    ;; Save the string around error place.
	    (setq string (gams-lst-save-string))
	    (setq file-name (gams-lst-get-input-filename))

	    ;; open GMS file.
	    (if (file-exists-p file-name)
		;;
		(progn (if (get-file-buffer file-name)
			   (switch-to-buffer (get-file-buffer file-name))
			 (find-file file-name))
		       ;; Search the error place.
		       (goto-char (point-min))
		       (if (re-search-forward string nil t)
			   (progn (recenter)
				  (move-to-column error-column))
			 ;;
			 (message "Error place is not found!")))
	      ;;
	      (message "The file `%s' does not exist!" file-name))))
      ;;
      (message "No error is found!"))))


(defun gams-lst-jump-to-input-file ()
  "Switch to the GMS file buffer."
  (interactive)
  (let ((file-gms (gams-lst-get-input-filename)))

    (if (null (file-exists-p file-gms))
	;; If gms file does not exist.
	(message "The file `%s' does not exist!" file-name)
      ;; If gms file exits.
      (if (get-file-buffer file-gms)
	  (switch-to-buffer (get-file-buffer file-gms))
	(find-file file-gms))
      (recenter))))

(defun gams-lst-jump-to-input-file-2 ()
  "Jump back to the error place in the input file.
"
  (interactive)
  (let ((file-gms (gams-lst-get-input-filename))
	string point-here)
    (forward-line 1)
    ;; Save the string around error place.
    (setq string (gams-lst-save-string))
    ;; open GMS file.
    (if (null (file-exists-p file-gms))
	;; If gms file does not exist.
	(message "The file `%s' does not exist!" file-name)
      ;; If gms file does exists.
      (if (get-file-buffer file-gms)
	  (switch-to-buffer (get-file-buffer file-gms))
	(find-file file-gms))
      
      (setq point-here (point))
      (if string
	  (progn
	    (goto-char (point-min))
	    ;; Search.
	    (save-excursion
	      (if (re-search-forward string nil t)
		  (progn
		    (recenter)
		    (beginning-of-line)
		    (setq point-here (point)))))))
      (goto-char point-here))))

;;;;

(defun gams-lst-jump-item (item &optional flag)
  "Jump to the next (or previous) ITEM (VAR, EQU, SUMMARY etc.)

If FLAG is non-nil, jump to the previous item.
"
  (let ((item-name item)
	(regex-sum "S O L V E      S U M M A R Y")
	(regex-rep "\\*\\*\\*\\* REPORT SUMMARY")
	(regex-var "^---- ")
	(regex-par "[0-9]+ PARAMETER "))
    (if (null flag)
	;; Jump to the next.
	(progn
	  (end-of-line)
	  (if (re-search-forward (cond
				  ((equal item "SUM") regex-sum)
				  ((equal item "REP") regex-rep)
				  ((equal item "PAR") regex-par)
				  (t (concat "^---- " item)))
				 nil t)
	      (progn
		(beginning-of-line)
		(recenter))
	    (message (concat "No more " item " entry"))))
      ;; Jump to the previous.
      (beginning-of-line)
      (if (re-search-backward (cond
			       ((equal item "SUM") regex-sum)
			       ((equal item "REP") regex-rep)
			       ((equal item "PAR") regex-par)
			       (t (concat "^---- " item)))
			      nil t)
	  (progn
	    (beginning-of-line)
	    (recenter))
	(message (concat "No more " item " entry"))))))

(defun gams-lst-solve-summary ()
  "Jump to the next SOLVE SUMMARY"
  (interactive)
  (interactive)
  (gams-lst-jump-item "SUM"))

(defun gams-lst-solve-summary-back ()
  "Jump to the previous SOLVE SUMMARY"
  (interactive)
  (gams-lst-jump-item "SUM" t))

(defun gams-lst-report-summary ()
  "Jump to the next REPORT SUMMARY"  
  (interactive)
  (gams-lst-jump-item "REP"))

(defun gams-lst-report-summary-back ()
  "Jump to the previous REPORT SUMMARY"
  (interactive)
  (gams-lst-jump-item "REP" t))

(defun gams-lst-next-var ()
  "Jump to the next VAR entry"
  (interactive)
  (gams-lst-jump-item "VAR"))

(defun gams-lst-previous-var ()
  "Jump to the previous VAR entry"
  (interactive)
  (gams-lst-jump-item "VAR" t))

(defun gams-lst-next-equ ()
  "Jump to the next EQU entry"
  (interactive)
  (gams-lst-jump-item "EQU"))

(defun gams-lst-previous-equ ()
  "Jump to the previous EQU entry"
  (interactive)
  (gams-lst-jump-item "EQU" t))

(defun gams-lst-next-par ()
  "Jump to the next PARAMETER entry"  
  (interactive)
  (gams-lst-jump-item "PAR"))

(defun gams-lst-previous-par ()
  "Jump to the next PARAMETER entry"  
  (interactive)
  (gams-lst-jump-item "PAR" t))

(defun gams-lst-widen-window ()
  "Make the window fill its frame.  Same as `delete-other-window'."  
  (interactive)
  (delete-other-windows)
  (recenter)
  (message "Winden window."))

(defun gams-lst-split-window ()
  "Split current window into two windows.  Same as `split-window-vertically'."  
  (interactive)
  (split-window-vertically)  
  (recenter)
  (message "Split window."))

(defun gams-lst-query-jump-to-line (line-num)
  "Jump to the line you specify."
  (interactive "sInput line number: ")
  (let (temp-num)
    (setq temp-num
	  (concat "^[ ]*" line-num))
    (goto-char (point-min))
    (re-search-forward temp-num nil t)
    (beginning-of-line)))

(defun gams-lst-jump-to-line ()
  "Jump to the line indicated by the number you are on.

If you execute this command on a line like

**** Exec Error 0 at line 32 .. Division by zero

you can jump to line 32."
  (interactive)
  (let ((cur-point (point))
	end-point
	line-num)
    ;;
    (end-of-line)
    (setq end-point (point))
    ;;	Get the line number.
    (beginning-of-line)
    (if (re-search-forward "at line \\([0-9]+\\)" end-point t)
	(progn
	  (setq line-num
		(concat "^[ ]*"
			(buffer-substring
			 (match-beginning 1)
			 (match-end 1))))
	  ;; Go to the beginning of the buffer
	  (goto-char (point-min))
	  ;; Search line.
	  (re-search-forward line-num nil t)
	  (beginning-of-line)
	  (message "If you want to jump to the GMS file, push `%s'."
		   gams-lk-4))
      ;;
      (goto-char cur-point)
      (message "This command is valid only if the cursor in on a line with line number!")
      )))

(defun gams-lst-move-cursor ()
  "Jump the cursor to the other window."
  (interactive)
  (other-window 1))

;; From the emasc lisp book written by Yuuji Hirose.
(defun gams-lst-resize-frame ()
  "Resize frame by keyboard.

n - Widen vertically
p - Narrow vertically
f - Widen horizontally
b - Narrow horizontally
Any other key - quit

To put Control key simultaneously makes movement faster.
"
  (interactive)
  (let (key
	(width (frame-width))
	(height (frame-height)))
    (catch 'quit
      (while t
	(message "Resize frame by [(C-)npfb] (%dx%d): " width height)
	(setq key (read-char))
	(cond
	 ((eq key ?n) (setq height (+ 1 height)))
	 ((eq key 14) (setq height (+ 5 height)))
	 ((eq key ?p) (setq height (- height 1)))
	 ((eq key 16) (setq height (- height 5)))
	 ((eq key ?f) (setq width (+ 1 width)))
	 ((eq key 6) (setq width (+ 5 width)))
	 ((eq key ?b) (setq width (- width 1)))
	 ((eq key 2) (setq width (- width 5)))
	 (t (throw 'quit t)))
	(modify-frame-parameters
	 nil (list (cons 'width width) (cons 'height height)))))
    (message "End...")))


;;; From the emacs lisp book written by Yuuji Hirose.
(defun gams-lst-move-frame ()
  "Move frame by keyboard.

n - Move upward
p - Move downward
f - Move rightward
b - Move leftward
Any other key - quit

To put Control key simultaneously makes movement faster.
"
  (interactive)
  (let (key
	(top (frame-parameter nil 'top))
	(left (frame-parameter nil 'left)))
    (if (listp top)
	(setq top (car (cdr top))))
    (if (listp left)
	(setq left (car (cdr left))))
    (catch 'quit
      (while t
	(message "Move frame by [(C-)npfb] (%dx%d): " top left)
	(setq key (read-char))
	(cond
	 ((eq key ?n) (setq top (+ 10 top)))
	 ((eq key ?p) (setq top (- top 10)))
	 ((eq key ?f) (setq left (+ 10 left)))
	 ((eq key ?b) (setq left (- left 10)))
	 ((eq key 14) (setq top (+ 20 top)))
	 ((eq key 16) (setq top (- top 20)))
	 ((eq key 6) (setq left (+ 20 left)))
	 ((eq key 2) (setq left (- left 20)))
	 (t (throw 'quit t)))
	(if (and (or (eq key ?p) (eq key 16)) (<= top 5))
	    (progn
;	      (message "Can't move upward anymore!")
	      (setq top 5)))
	(if (and (or (eq key ?b) (eq key 2)) (<= left 5))
	    (progn
;	      (message "Can't move left anymore!")
	      (setq left 5)))
	(modify-frame-parameters
	 nil (list (cons 'top top) (cons 'left left )))))
    (message "End...")))


;;; scroll
(defun gams-lst-count-win ()
  "Count the number of windows and return it."
  (let ((cur-win (selected-window))
	(num 1)
	flag)
    (save-excursion
      (if (one-window-p)
	  nil
	;;
	(catch 'quit
	  (while t
	    (other-window 1)
	    (setq flag (selected-window))
	    (cond
	     ((not (eq flag cur-win))
	      (setq num (+ num 1)))
	     ((eq flag cur-win)
	      (throw 'quit t)))))))
    num))

(defun gams-lst-scroll (&optional down num page)
  "Command for scrolling.

If DOWN is non-nil, scroll down.
NUM mean scroll type (nil, 2, or d).
If PAGE is non-nil, page scroll."
  (interactive)
  (let ((cur-win (selected-window))
	(win-num (gams-ol-count-win))
	;; flag for lst or ov?
	(flag-lst
	 (if (equal "GAMS-LST" mode-name) t nil))
	;; flag for page scroll or not.
	(fl-pa (if page nil 1)))
		      
    (if flag-lst
	;; If LST buffer
	(cond
	 ;; scroll type 1.
	 ((null num)
	  (save-excursion
	    (if down (scroll-down fl-pa)
	      (scroll-up fl-pa))))
	 ;; scroll type 2.
	 ((equal num "2")
	  (cond
	   ((eq win-num 1)
	    (if down
		(scroll-down fl-pa)
	      (scroll-up fl-pa)))
	   ((> win-num 1)
	    (other-window 1)
	    (if down
		(scroll-down fl-pa)
	      (scroll-up fl-pa))
	    (select-window cur-win))
	   (t nil)))
	 ;; scroll type double.
	 ((equal num "d")
	  (cond
	   ((eq win-num 1)
	    (if down
		(scroll-down fl-pa)
	      (scroll-up fl-pa)))
	   ((> win-num 1)
	    (if down
		(scroll-down fl-pa)
	      (scroll-up fl-pa))
	    (other-window 1)
	    (if down
		(scroll-down fl-pa)
	      (scroll-up fl-pa))
	    (select-window cur-win))
	   (t nil))))

      ;; If OV buffer.
      (cond
       ;; scroll type 1.
       ((null num)
	(cond
	 ((eq win-num 1)
	  nil)
	 ((> win-num 1)
	  (save-excursion
	    (other-window 1)
	    (if down
		(scroll-down fl-pa)
	      (scroll-up fl-pa))
	    (select-window cur-win)))))
       ;; scroll type 2.
       ((equal num "2")
	(cond
	 ((eq win-num 1)
	  nil)
	 ((eq win-num 2)
	  (other-window 1)
	  (if down
	      (scroll-down fl-pa)
	    (scroll-up fl-pa))
	  (select-window cur-win))
	 ((eq win-num 3)
	  (other-window 2)
	  (if down
	      (scroll-down fl-pa)
	    (scroll-up fl-pa))
	  (select-window cur-win))
	 (t nil)))
       ;; scroll type double.
       ((equal num "d")
	(cond
	 ((eq win-num 1)
	  nil)
	 ((eq win-num 2)
	  (other-window 1)
	  (if down
	      (scroll-down fl-pa)
	    (scroll-up fl-pa))
	  (select-window cur-win))
	 ((eq win-num 3)
	  (other-window 1)
	  (if down
	      (scroll-down fl-pa)
	    (scroll-up fl-pa))
	  (other-window 1)
	  (if down
	      (scroll-down fl-pa)
	    (scroll-up fl-pa))
	  (select-window cur-win))
	 (t nil))))
      ;;
      (gams-ol-show-key)
      )))

;;; line scroll.
(defun gams-lst-scroll-1 ()
  (interactive)
  (gams-lst-scroll))

(defun gams-lst-scroll-down-1 ()
  (interactive)
  (gams-lst-scroll t))

(defun gams-lst-scroll-2 ()
  (interactive)
  (gams-lst-scroll nil "2"))

(defun gams-lst-scroll-down-2 ()
  (interactive)
  (gams-lst-scroll t "2"))

(defun gams-lst-scroll-double ()
  (interactive)
  (gams-lst-scroll nil "d"))

(defun gams-lst-scroll-down-double ()
  (interactive)
  (gams-lst-scroll t "d"))

;;; Page scroll
(defun gams-lst-scroll-page-1 ()
  (interactive)
  (gams-lst-scroll nil nil t))

(defun gams-lst-scroll-page-down-1 ()
  (interactive)
  (gams-lst-scroll t nil t))

(defun gams-lst-scroll-page-2 ()
  (interactive)
  (gams-lst-scroll nil "2" t))

(defun gams-lst-scroll-page-down-2 ()
  (interactive)
  (gams-lst-scroll t "2" t))

(defun gams-lst-scroll-page-double ()
  (interactive)
  (gams-lst-scroll nil "d" t))

(defun gams-lst-scroll-page-down-double ()
  (interactive)
  (gams-lst-scroll t "d" t))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;	Code for GAMS-TEMPLATE.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(autoload 'gams-template "gams-temp" "Start GAMS-TEMPLATE" t)

;;; Customizable variables for GAMS-TEMPLATE mode.

(defcustom gams-template-file "~/.gams-template"
  "*The name of a file used to store templates."
  :type 'file
  :group 'gams)
  
(defcustom gams-save-template-change nil
  "Nil means that save the content of `gams-user-template-alist' into
`gams-template-file' only when you quit Emacs.  If non-nil, save
`gams-user-template-alist' every time you made changes.  If your Emacs
often crashes, you may had better set it to non-nil."
  :type 'boolean
  :group 'gams)

(defcustom gams-template-mark "%c"
  "*The mark that indicates the point of cursor in a template."
  :type 'string
  :group 'gams)

;;; Customizable variables end here.


;;;;; Codes

(if (not gams-save-template-change)
    (add-hook 'kill-emacs-hook 'gams-write-alist-to-file))

(if (file-exists-p gams-template-file)
    (progn
      (load-file gams-template-file)
      (defvar gams-user-template-alist-init gams-user-template-alist)))

(defvar gams-user-template-alist nil)
(defvar gams-temp-buffer "*Template List*")
(defvar gams-temp-edit-buffer "*Template Edit*")
(defvar gams-temp-cont-buffer "*Template Content*")
(defvar gams-prog-file-buff nil)

(defun gams-template ()
  "Start the GAMS-TEMPLATE mode."
  (interactive)
  (let* ((temp-buffer (get-buffer-create gams-temp-buffer))
	 (cur-buf (current-buffer)))

    ;; store window configuration.
    (setq gams-temp-window (current-window-configuration))
    
    (save-excursion
      (pop-to-buffer temp-buffer)
      (gams-template-mode cur-buf)
      )))

;; Key assignment of GAMS-TEMPLATE mode.
(setq gams-template-mode-map (make-keymap))
(let
    ((map gams-template-mode-map))
  (define-key map " "     'gams-temp-show-cont)
  (define-key map "\r"    'gams-temp-exit)
  (define-key map "q"     'gams-temp-quit)
  (define-key map "a"     'gams-temp-add)

  (define-key map "d"     'gams-temp-delete)
  (define-key map "e"     'gams-temp-reedit)
  (define-key map "r"     'gams-temp-rename)
  (define-key map "u"     'gams-temp-up)
  (define-key map "j"     'gams-temp-down)
  (define-key map "g"     'gams-temp-gms)
  (define-key map "s"     'gams-temp-scroll)
  (define-key map "S"     'gams-temp-scdown)
  (define-key map "h"     'gams-temp-help)
  (define-key map "o"     'gams-write-alist-to-file)

  (define-key map "p"     'gams-temp-prev)
  (define-key map "n"     'gams-temp-next)
  (define-key map [up]     'gams-temp-prev)
  (define-key map [down]     'gams-temp-next)
  )

;;; Menu for GAMS-TEMPLATE mode.
(easy-menu-define 
  gams-template-mode-menu gams-template-mode-map "Menu keymap for GAMS-TEMPLATE mode."
  '("GAMS-TEMPLATE"
    ["Show a content of a template" gams-temp-show-cont t]
    ["Insert a template" gams-temp-exit t]
    ["Quit TEMPLATE mode" gams-temp-quit t]
    ["Add a new template" gams-temp-add t]
    "--"    
    ["Delete a template" gams-temp-delete t]
    ["Re-edit a template" gams-temp-reedit t]
    ["Rename a template" gams-temp-rename t]
    ["Move a template up" gams-temp-up t]
    ["Move a template down" gams-temp-down t]
    ["Show gms file" gams-temp-gms t]
    "--"
    ["Scroll up Content buffer" gams-temp-scroll t]
    ["Scroll down Content buffer" gams-temp-scdown t]
    "--"
    ["Show help" gams-temp-help t]
    ["Save" gams-write-alist-to-file t]
    "--"
    ["Show the previous template" gams-temp-prev t]
    ["Show the next template" gams-temp-next t]
    ))

;; GAMS-TEMPLATE mode.
(defun gams-template-mode (&optional buff-name)
  "The GAMS-TEMPLATE mode (a mode for template handling).

BUFF-NAME is the name of the current LST file buffer.

The following commands are available in this mode.

\\[gams-temp-show-cont]		Show a content of a template.
\\[gams-temp-exit]		Insert a template.
\\[gams-temp-quit]		Quit.
\\[gams-temp-add]		Add a new template.

\\[gams-temp-delete]		Delete a template.
\\[gams-temp-reedit]		Re-edit a template.
\\[gams-temp-rename]		Rename a template.
\\[gams-temp-up]		Move a template up.
\\[gams-temp-down]		Move a template down.

\\[gams-temp-gms]		Show the gms file.
\\[gams-temp-scroll](\\[gams-temp-scdown])		Scroll up (down) *Template Content* buffer.
\\[gams-temp-help]		Show this help.
\\[gams-write-alist-to-file]		Save the content of gams-user-template-alist.

\\[gams-temp-prev]		Show the previous template.
\\[gams-temp-next]		Show the next template.
"
  (let ((temp-buff gams-prog-file-buff))
    (kill-all-local-variables)
    (setq mode-name "GAMS-TEMPLATE"
	  major-mode 'gams-template-mode)
    (use-local-map gams-template-mode-map)

    ;; Make a buffer local variable.
    (if (null temp-buff)
	(progn
	  (make-local-variable 'gams-prog-file-buff)
	  (setq gams-prog-file-buff buff-name))
      (make-local-variable 'gams-prog-file-buff)
      (setq gams-prog-file-buff temp-buff))
    
    (setq buffer-read-only nil)

    (gams-temp-show-list)
    (gams-temp-show-cont)

    (if gams-user-template-alist
	(gams-temp-select-key)
      (gams-temp-show-message))
    
    (setq buffer-read-only t)

    ;; menu.
    (easy-menu-add gams-template-mode-menu)
    ))

(defun gams-temp-select-key ()
  "Show key assignments in the GAMS-TEMPLATE mode."
  (message
   (format "[h]elp, [p]rev, [n]ext, SPACE = show, [g]ms, [q]uit, RET = insert, [a]dd. ")))

(defun gams-temp-show-list ()
  "Insert template list in the *Template List* buffer."
  (save-excursion
    (let ((temp-alist gams-user-template-alist)
	  temp-ele
	  temp-ele-name
	  temp-ele-cont
	  (buffer-read-only nil))

      (erase-buffer)
      (goto-char (point-min))

      (if temp-alist
	  (progn
	    (while temp-alist
	      (setq temp-ele (car temp-alist))
	      (setq temp-ele-name (car temp-ele))
	      (setq temp-ele-cont (cdr temp-ele))
	      
	      (beginning-of-line)
	      (move-to-column 2 t) 
	      (insert
	       (concat "[" temp-ele-name "]"))
	      (insert "\n")
	      
	      (setq temp-alist (cdr temp-alist))
	      )
	    ;; Narrow the region.
	    (narrow-to-region (point-min)
			      (- (point-max) 1))
	    (goto-char (point-min)))
	
	(gams-temp-show-message)))))


(defun gams-temp-show-message ()
  (message
   (concat "No template is registered!"
	   "  `a' = add, `q' = quit, `h' = help.")))

(defun gams-temp-show-cont ()
  "Show the content of a template in the *Template Content* buffer. "
  (interactive)
  (let ((curr-buf (current-buffer))
	(temp-buf (get-buffer-create gams-temp-cont-buffer))
	(temp-na (gams-temp-get-name)))
    (if gams-user-template-alist
	(if temp-na
	    (progn
	      (pop-to-buffer temp-buf)
	      (let ((buffer-read-only nil))
		(erase-buffer)
		(goto-char (point-min))
		(insert (cdr (assoc temp-na gams-user-template-alist)))
		(setq template (assoc temp-na gams-user-template-alist))
		;; Change the Mode.
		(gams-temp-cont-mode)
		(goto-char (point-min))
		(pop-to-buffer curr-buf)
		(setq buffer-read-only t)
		(gams-temp-select-key)
		)))
      (gams-temp-show-message)
      )))

(defun gams-temp-cont-mode ()
  "The mode for *Template Content* buffer."
  (kill-all-local-variables)
  (setq major-mode 'gams-temp-cont-mode)
  (setq mode-name "Content")
  ;; Font-lock
;   (make-local-variable 'font-lock-defaults)
;   (setq font-lock-defaults '(gams-font-lock-keywords t t))
  (setq buffer-read-only t)
  )

(defun gams-temp-get-name ()
  "Get a name of a template on the current line."
  (interactive)
  (save-excursion
    (let (point-a point-b temp-na)
      (end-of-line)
      (setq point-a (point))
      (beginning-of-line)
      (if (re-search-forward "[   ]+[[]\\(.*\\)[]]" point-a t)
	  (progn
	    (setq temp-na (buffer-substring (match-beginning 1)
					    (match-end 1)))
	    temp-na)))))

(defun gams-temp-scroll ()
  "Scroll up *Template Content* buffer."
  (interactive)
  (scroll-other-window 1)
  (gams-temp-select-key)
  (interactive))
  
(defun gams-temp-scdown ()
  "Scroll down *Template Content* buffer."
  (interactive)
  (scroll-other-window-down 1)
  (gams-temp-select-key)
  (interactive))
  
(defun gams-temp-help ()
  "Display help for GAMS-TEMPLATE mode."
  (interactive)
  (describe-function 'gams-template-mode))


(defun gams-temp-quit ()
  "Quit the GAMS-TEMPLATE mode."
  (interactive)
  (if gams-save-template-change
      (gams-write-alist-to-file))
  (pop-to-buffer gams-prog-file-buff)
  (kill-buffer gams-temp-cont-buffer)
  (kill-buffer gams-temp-buffer)
  (delete-other-windows)
  ;; restore window configurations.
  (set-window-configuration gams-temp-window))

(defun gams-temp-next ()
  "Display the next template."
  (interactive)
  (let
      ((sig-max-num (length gams-user-template-alist)))
    (if gams-user-template-alist
	(progn
	  (if (null (equal sig-max-num (count-lines (point-min) (+ 1 (point)))))
	      (next-line 1))
	  (gams-temp-show-cont)
	  (setq buffer-read-only t))
      (gams-temp-show-message))))

(defun gams-temp-prev ()
  "Display the previous template."
  (interactive)
  (if gams-user-template-alist
      (progn
	(if (null (equal 1 (count-lines (point-min) (+ (point) 1))))
	    (next-line -1))
	(gams-temp-show-cont)
	(setq buffer-read-only t))
    (gams-temp-show-message)))

(defun gams-temp-gms ()
  "Show the gms file."
  (interactive)
  (let ((temp-buf (current-buffer)))
  ;;
  (pop-to-buffer gams-prog-file-buff)
  (pop-to-buffer temp-buf)
  (gams-temp-select-key)))

(defun gams-temp-internal (temp)
  (let (point-a)
    (if temp
	(if (string= temp "")
	    nil
	  (narrow-to-region (point-min)
			    (setq point-a (point)))
	  (insert temp)
	  (gams-temp-replace point-a (point-max))
	  (widen)
	(set-buffer-modified-p (buffer-modified-p))))))

(defun gams-temp-exit ()
  "Insert a template into a buffer."
  (interactive)
  (let*
      ((temp-name (gams-temp-get-name))
       (temp-cont (cdr (assoc temp-name gams-user-template-alist))))

    ;; Back to the program file buffer.
    (switch-to-buffer gams-prog-file-buff)
    (delete-other-windows)
    ;; Insert a template.
    (gams-temp-internal temp-cont)
    (kill-buffer gams-temp-cont-buffer)
    (kill-buffer gams-temp-buffer)
    ;; restore window configurations.
    (set-window-configuration gams-temp-window)))

(defun gams-temp-add ()
  "Add a new template."
  (interactive)
  ;; kill template buffers.
  (kill-buffer gams-temp-cont-buffer)
  (gams-edit-template))

(defun gams-temp-reedit ()
  "Re-edit already registered templates."
  (interactive)
  (let
      ((temp-name (gams-temp-get-name))
       (temp-line (count-lines (point-min) (+ (point) 1)))
       key)
    ;;
    (gams-temp-show-cont)
    ;; Go to "*Template Content*" buffer.
    (switch-to-buffer gams-prog-file-buff)
    (pop-to-buffer (get-buffer-create gams-temp-cont-buffer))
    (if (get-buffer gams-temp-edit-buffer)
	(kill-buffer gams-temp-edit-buffer))
    (rename-buffer gams-temp-edit-buffer)
    ;; Switch to gams-edit-template mode.
    (gams-edit-template temp-name)
    (switch-to-buffer gams-temp-buffer)
    (goto-line temp-line)
    (gams-temp-show-cont)))

(defun gams-temp-rename ()
  "Rename already registered templates."
  (interactive)
  (let*
      ((buffer-read-only nil)
       (temp-alist gams-user-template-alist)
       (old-name (gams-temp-get-name))
       (new-name
	(read-input "Insert a new name: "
		    old-name))
       (line-num (count-lines (point-min)
				   (+ 1 (point)))))
    (if (null temp-alist)
	nil
      ;; Replace the old name with the new name.
      (setcar (assoc old-name temp-alist) new-name)
      ;; If gams-save-template-change is non-nil, save.
      (if gams-save-template-change
	  (gams-write-alist-to-file))
      ;;
      (gams-temp-show-list)
      (goto-line line-num)
      (gams-temp-show-cont))))

(defun gams-temp-delete ()
  "Delete existing templates."
  (interactive)
  (if gams-user-template-alist
      (progn
	(let
	    ((temp-name (gams-temp-get-name))
	     (line-num (count-lines (point-min)
				    (+ 1 (point))))
	     key)
	  ;; Delete.
	  (message
	   (format "Do you really want to delete `%s'?  Type `y' if yes." temp-name))
	  (setq key (read-char))
	  (if (equal key ?y)
	      (progn
		(forward-line 1)
		(gams-template-processing "del" temp-name)
		(gams-temp-show-list)
		(goto-line (- line-num 1))
		(gams-temp-show-cont))
	    )))))

(defun gams-temp-up ()
  "Move up a template."
  (interactive)
  (if gams-user-template-alist
      (progn
	(let ((temp-name (gams-temp-get-name))
	      (temp-alist gams-user-template-alist)
	      (line-num (count-lines (point-min)
				     (+ 1 (point)))))
	  (if (equal 1 line-num)
	      nil
	    ;;
	    (setq gams-user-template-alist
		  (gams-temp-alist-change temp-alist temp-name)))
	  (gams-temp-show-list)
	  (goto-line (- line-num 1))
	  (gams-temp-show-cont)
	  ))))

(defun gams-temp-down ()
  "Move down a template."
  (interactive)
  (if gams-user-template-alist
      (progn
	(let ((temp-name (gams-temp-get-name))
	      (temp-alist gams-user-template-alist)
	      (line-num (count-lines (point-min)
				     (+ 1 (point))))
	      (sig-max-num (length gams-user-template-alist)))
	  (if (equal sig-max-num line-num)
	      nil
	    ;;
	    (setq gams-user-template-alist
		  (gams-temp-alist-change temp-alist temp-name t)))
	  (gams-temp-show-list)
	  (goto-line (+ line-num 1))
	  (gams-temp-show-cont)
	  ))))

;; Editing templates.
(defun gams-temp-add-key ()
  (message
   (format "Edit a template!  C-xC-s = save and exit, C-cC-s = save, C-xk = quit, C-xh = help")))

; key assignment.
(defvar gams-template-edit-map (make-keymap) "keymap for gams-template-edit")
(let ((map gams-template-edit-map))
  (define-key map "\C-ck" 'gams-insert-statement)
  (define-key map "$" 'gams-insert-dollar-control)
  (define-key map "\C-cd" 'gams-insert-dollar-control)
  (define-key map "(" 'gams-insert-parens)
  (define-key map "\"" 'gams-insert-double-quotation)
  (define-key map "`" 'gams-insert-single-quotation)

  (define-key map "\C-co" 'gams-insert-comment)
  (define-key map "\t" 'gams-insert-tab)

  (define-key map "\C-xh" 'gams-edit-template-help)
  (define-key map "\C-x\C-s" 'gams-save-template)
  (define-key map "\C-c\C-s" 'gams-save-and-exit-template)
  (define-key map "\C-xk" 'gams-quit-template))

;;; Menu for GAMS-TEMPLATE-EDIT mode. 
(easy-menu-define 
  gams-template-edit-menu gams-template-edit-map "Menu keymap for GAMS-TEMPLATE-EDIT mode."
  '("TEMPLATE-EDIT"
    ["Save the template" gams-save-template t]
    ["Save the template and exit" gams-save-and-exit-template t]
    ["Show help" gams-edit-template-help t]
    ["Quit TEMPLATE-EDIT mode" gams-quit-template t]
    "--"
    ["Insert GAMS statement" gams-insert-statement t]
    ["Insert GAMS dollar control" gams-insert-dollar-control t]
    ["Insert parenthesis" gams-insert-parens t]
    ["Insert double quotations" gams-insert-double-quotation t]
    ["Insert single quotations" gams-insert-single-quotation t]
    ["Insert a comment template" gams-insert-comment t]
    ))

(defun gams-edit-template (&optional file)
  "Edit a template.

FILE is a file name.  It is used for gams-temp-reedit.

'\\[gams-save-template] - Save a template.
'\\[gams-save-and-exit-template] - Save a template and exit.
'\\[gams-quit-template] - Quit.
'\\[gams-edit-template-help] - Show this help.

'\\[gams-insert-statement] - Insert GAMS statement with completion.
'\\[gams-insert-dollar-control] - Insert GAMS statement (dollar control option).
'\\[gams-insert-parens] - Insert parenthesis.
'\\[gams-insert-double-quotation] - Insert double quotations.
'\\[gams-insert-single-quotation] - Insert single quotations.

'\\[gams-insert-comment] - Insert comment template.

"
  (interactive)
  (delete-other-windows)
  (pop-to-buffer gams-temp-edit-buffer)
  
  (kill-all-local-variables)
  (setq major-mode 'gams-template-edit)
  (setq mode-name "TEMPLATE-EDIT")
  (use-local-map gams-template-edit-map)
  (setq fill-prefix "\t\t")
  (mapcar
   'make-local-variable
   '(fill-column
     fill-prefix
;      paragraph-start
;      paragraph-separate
     indent-line-function
     comment-start
     comment-start-skip
	    ))
  (setq
   fill-column gams-fill-column
   fill-prefix gams-fill-prefix
;    paragraph-start    gams-paragraph-start
;    paragraph-separate gams-paragraph-separate
   indent-line-function 'gams-indent-line-function
   comment-start gams-comment-prefix
   comment-end ""
   comment-start-skip "^*+[ \t]*"
   )

  ;; Create the alist of statements.  Is this necessary?
  ;; See `gams-statement-update'.
  (setq gams-statement-alist
	(gams-statement-to-alist gams-statement-up gams-statement-upcase))
  
  ;; Create the alist of dollar control options.  Is this necessary?
  ;; See `gams-statement-update'.  
  (if gams-use-mpsge
      ;; Use mpsge.
      (progn
	(setq gams-dollar-control-alist
	      (gams-statement-to-alist
	       (append gams-dollar-control-up gams-statement-mpsge)
	       gams-dollar-control-upcase)))
    ;; Not use mpsge
    (setq gams-dollar-control-alist
	  (gams-statement-to-alist gams-dollar-control-up gams-dollar-control-upcase)))
  ;; Font-lock
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(gams-font-lock-keywords t t))
  ;; 
  (if (file-exists-p gams-statement-file)
      (load-file gams-statement-file))
  ;; Update.
  (gams-statement-update)
  ;; set file name used for 
  (make-local-variable 'gams-add-template-file)
  (setq gams-add-template-file file)
  (setq buffer-read-only nil)
  ;; TEST.
  (set-buffer-modified-p nil)

  (current-buffer)
  (gams-temp-add-key)

  ;;
  (easy-menu-add gams-template-edit-menu)

  (recursive-edit)

  (kill-buffer gams-temp-edit-buffer)    
  (switch-to-buffer gams-temp-buffer)
  (gams-temp-show-list)
  (gams-temp-show-cont)
  );; gams-edit-template ends here.

(defun gams-edit-template-help ()
  "Show help."
  (interactive)
  (describe-function 'gams-edit-template))

(defun gams-save-and-exit-template ()
  ""
  (interactive)
  (gams-save-template)
  (exit-recursive-edit))

(defun gams-save-template ()
  "Register a template."
  (interactive)
  (save-excursion
    (let* ((temp-name (read-input
		       "Enter a name of this template: "
		       gams-add-template-file))
	   (temp-alist gams-user-template-alist)
	   list-tmp over-flag)
      (if (string= temp-name "")
	  (error "Need to specify a name of this template!")
	(setq list-tmp
	      (list temp-name
		    (buffer-substring (point-min) (point-max))))
	;; The same name is already used?
	(if (assoc temp-name temp-alist)
	    ;; Already used.
	    
	    ;; Overwrite it?
	    (if (setq over-flag
		      (y-or-n-p
		       "This template name is already exists. Do you want to override it?: "))
		;; Yes
		(gams-template-processing "red"
					  (car list-tmp)
					  (car (cdr list-tmp)))
	      ;; No. Do nothing.
	      nil)
	  ;; The same name is not registered. 
	  (gams-template-processing "reg"
				    (car list-tmp)
				    (car (cdr list-tmp)))
	  )))))
  
(defun gams-quit-template ()
  ""
  (interactive)
  (exit-recursive-edit))

(defun gams-search-default-temp ()
  (let ((ret nil)
	(functions sig-search-functions))
    (while functions
      (if (setq ret (funcall (car functions)))
	  (setq functions nil)
	(setq functions (cdr functions))))
    (or ret sig-default-name)))

(defun gams-register-template-edit (tem-cont)
  "Escape double quoatation and backslash."
  (let
      ((temp-buff "*gams temp buff")
       string)
    (save-excursion
      (get-buffer-create temp-buff)
      (switch-to-buffer temp-buff) ;; Just for debug.
      (erase-buffer)
      (goto-char (point-min))
      (insert tem-cont)

      ;; Replace several strings with regular expression.
      (goto-char (point-min))
      (while (re-search-forward "\\(\\\\\\)" nil t)
	(setq a-point (match-beginning 1))
	(setq b-point (match-end 1))
	(replace-match "\\\\\\\\"))
      (goto-char (point-min))
      (while (re-search-forward "\\(\"\\)" nil t)
	(setq a-point (match-beginning 1))
	(setq b-point (match-end 1))
	(replace-match "\\\\\""))
      ;; Store regular expression to string.
      (setq string (buffer-substring (point-min) (point-max)))
      ;; Delete temporary buffer.
      (kill-buffer temp-buff)
      ;; Return the content of string.
      string)))

(defun gams-temp-write-alist ()
  "Update `gams-user-template-alist'."
  (let
      ((temp-list gams-user-template-alist))
    (erase-buffer)
    (insert-string
     (concat
      "(setq gams-user-template-alist '(\n"))
    (goto-char (point-max))
    (while temp-list
      (insert
       (concat
	"(\"" (car (car temp-list)) "\" . \""
	(gams-register-template-edit
	 (cdr (car temp-list)))
	"\")\n"))
      (goto-char (point-max))
      (setq temp-list (cdr temp-list)))
    (insert "))\n")
    ;;
    (eval-buffer)))

(defun gams-template-processing (type name &optional cont)
  "Process a template in a temporary buffer.

TYPE is a sting.
reg = register a new template.
del = delete.
red = re-edit."
  (let ((cur-buff (current-buffer))
	(temp-buff " *gams-temporary*")
	(temp-file gams-template-file)
	temp-cont)
    ;; Move to the temporary buffer
    (get-buffer-create temp-buff)
    (switch-to-buffer temp-buff)
    ;; Cases:
    (cond
    ;; register
     ((equal type "reg")
      (setq gams-user-template-alist
	    (append (list (cons name cont))
		    gams-user-template-alist)))
    ;; delete
     ((equal type "del")
      (setq gams-user-template-alist
	    (delete
	     (assoc name gams-user-template-alist)
	     gams-user-template-alist)))
    ;; red
     ((equal type "red")
      (setcdr (assoc name gams-user-template-alist) cont)))
    ;; reedit. not yet.

    ;; Write the content of gams-user-template-alist in the temporary
    ;; buffer.
    (gams-temp-write-alist)
    
    ;; If gams-save-template-change is non-nil, save the temporary buffer
    ;; as gams-template-file.
    (if gams-save-template-change
	(write-file gams-template-file))
    (kill-buffer (get-file-buffer gams-template-file))
    (switch-to-buffer cur-buff)
    ))

(defun gams-write-alist-to-file ()
  "Save the content of `gams-user-template-alist' into the file
`gams-user-template-alist'."
  (interactive)
  (save-excursion
    (if gams-user-template-alist
	(if (null
	     (eq gams-user-template-alist gams-user-template-alist-init))
		  (progn
		    (set-buffer (get-buffer-create " *gams-temporary*"))
		    (gams-temp-write-alist)
		    (write-file gams-template-file)
		    (kill-buffer (get-file-buffer gams-template-file)))))))
  
(defun gams-temp-alist-change (alist ele &optional flag)
  "Reorder `gams-user-template-alist'.

ELE is car part.
If FLAG is t, move down."
  (interactive)
  (let ((temp-alist alist)
	(car-p ele)
	alist-a
	alist-b
	list-a
	list-b)
    ;; Judge whether alist include ele.
    (if (setq alist-b
	      (member (assoc car-p temp-alist)
		      temp-alist))
	;; Included.
	(progn
	  (cond
	  ;; If flag is nil
	   ((null flag)
	    ;; If ele is the first element, do nothing.
	    (if (null (equal temp-alist alist-b))
		;; If ele is the second or later element 
		(progn
		  ;; 
		  (setq list-a (car alist-b))
		  (setq alist-b (cdr alist-b))
		  ;;
		  (setq temp-alist (reverse temp-alist))
		  (setq alist-a
			(member (assoc car-p temp-alist)
				temp-alist))
		  (setq alist-a	(cdr alist-a))
		  (setq list-b (car alist-a))
		  (setq alist-a (cdr alist-a))
		  (setq alist-a (append (list list-a) alist-a))
		  (setq alist-a (reverse alist-a))
		  (setq alist-b (append (list list-b) alist-b))
		  (setq temp-alist (append alist-a alist-b))
		  )))
	   ;; If flag is t
	   (flag
	    ;; If ele is the last element, do nothing.
	    (if	(null (equal alist-b (nth (- (length temp-alist) 1) temp-alist)))
		(progn
		  ;; alist-b
		  (setq list-a (car alist-b))
		  (setq list-b (car (cdr alist-b)))
		  (setq alist-b (cdr (cdr alist-b)))
		  (setq alist-b (append (list list-a) alist-b))
		  (setq alist-b	(append (list list-b) alist-b))
		  ;; alist-a
		  (setq temp-alist (reverse temp-alist))
		  (setq alist-a	(member (assoc car-p temp-alist) temp-alist))
		  (setq alist-a	(cdr alist-a))
		  (setq alist-a (reverse alist-a))
		  ;; New alist.
		  (setq temp-alist (append alist-a alist-b))
		  ))))))
    temp-alist))

(defun gams-temp-replace (begin end)
  "Replace a mark with cursor in the inserted template.

BEGIN and END are points."
  (interactive)
  (let ((mark gams-template-mark))
    (goto-char begin)
    (if (re-search-forward (format "\\(%s\\)" mark) end t)
	(progn
	  (goto-char (match-beginning 0))
	  (replace-match ""))
      (goto-char end))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;	Code for the GAMS-OUTLINE mode.
;;;	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;(autoload 'gams-outline "gams-ol" "Start GAMS-OUTLINE" t)

;;; Variables for GAMS-OUTLINE mode.

(defcustom gams-ol-height 15
  "*The height of the OUTLINE buffer with one LST buffer."
  :type 'number
  :group 'gams)

(defcustom gams-ol-height-two 8
  "*The height of the OUTLINE buffer with two LST buffers."
  :type 'number
  :group 'gams)

(defcustom gams-ol-view-item
      '(("SUM" . t)
	("VAR" . t)
	("EQU" . t)
	("PAR" . t)
	("SET" . t)
	("VRI" . t)
	("LOO" . t)
	("OTH" . t)
	("COM" . t))
  "The default alist of viewable items.

Each list consists of a pair of the item name and its flag.  Non-nil of
flag means the item is viewable by default.

The order of items has the meaning in this alist.  Items are listed in the
SELECT-ITEM buffer according to this order.  So, if you want to show MAR
on the top, you must write MAR at the fisrt in this alist.
"
  :type '(repeat (cons :tag "option" (string :tag "item") (boolean :tag "flag")))
  :group 'gams)

;;; Key for gams-outline.

(defcustom gams-olk-1 "?"
  "*Key for `gams-ol-help'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-olk-2 "N"
  "*Key for `gams-ol-view-next'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-olk-3 "P"
  "*Key for `gams-ol-view-prev'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-olk-4 "t"
  "*Key for `gams-ol-select-item'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-olk-5 " "
  "*Key for `gams-ol-view-base'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-olk-6 "q"
  "*Key for `gams-ol-quit'."
  :type 'string
  :group 'gams-keys)

(defcustom gams-olk-7 "m"
  "*Key for `gams-ol-mark'."
  :type 'string
  :group 'gams-keys)

;;; Codes for outlineing LST files.  

(defun gams-ol-create-list (alist)
  "Create a list of car part from ALIST."
  (let ((temp-alist (reverse alist))
	new-list)
    (while temp-alist
      (setq new-list (append (list (car (car temp-alist))) new-list))
      (setq temp-alist (cdr temp-alist)))
    new-list))

(setq gams-ol-item
      (gams-ol-create-list gams-ol-view-item))
      
(defun gams-ol-store-point (&optional flag)
  "Store points of re-search-forward results.

If FLAG is non-nil, it means the item is VARIABLE."
  (let
      (par-type2
       par-name2
       par-exp2)
    (if flag
	(setq par-type2 "VRI")
      (setq  par-type2
	     (substring (upcase
			 (buffer-substring
			  (match-beginning 1)
			  (match-end 1)))
			0 3)))
    (setq par-name2 (buffer-substring
		      (match-beginning 2)
		      (match-end 2)))
    (setq par-exp2 (buffer-substring
			 (match-beginning 3)
			 (match-end 3)))
    (list par-type2 par-name2 par-exp2)))

  
(defun gams-outline ()
  "Evoke the GAMS-OUTLINE mode."
  (interactive)
  (let ((buf-f-name (buffer-file-name))
	(buffname (concat "*"
			  (file-name-sans-extension
			   (file-name-nondirectory
			    (buffer-file-name))) "-OUTLINE*"))
	(lst-file-buf (current-buffer)))
    ;; Judge whether OV buffer is already created or not.
    (if (or (null (get-buffer buffname))
	    (null gams-ol-flag))
	;; OV buffer is not created or out of date.
	(progn
	  (setq gams-ol-flag t)
	  (get-buffer-create buffname)
	  (gams-ol-show buffname
			(gams-ol-convert (gams-ol-make-alist)))))
    ;; Switch to the OV buffer.
    (switch-to-buffer buffname)
    (recenter)
    ;; Start gams-ol-mode.
    (gams-ol-mode lst-file-buf)
    ))

(defun gams-ol-get-alist (&optional buffer)
  "Return the value of `gams-over-alist' in the current LST buffer.

If BUFFER is non-nil, the current buffer is OV buffer, not LST buffer."
  (interactive)
  (let (alist
	(cur-buf (if buffer (current-buffer) nil)))
    ;; The current buffer is LST buffer or OV buffer?
    (if cur-buf
	;; If OV buffer, switch to LST buffer.
	(progn
	  (set-buffer gams-ol-lst)
	  (setq alist gams-over-alist)
	  (switch-to-buffer cur-buf))
      ;; LST buffer.
      (setq alist gams-over-alist))
    ;; Return the value of alist.
    alist))
    
; key assignment.
(defvar gams-ol-mode-map (make-keymap) "Keymap for GAMS-OUTLINE mode")
(let ((map gams-ol-mode-map))
  (define-key map gams-olk-5 'gams-ol-view-base)
  (define-key map gams-olk-2 'gams-ol-view-next)
  (define-key map gams-olk-3 'gams-ol-view-prev)
  (define-key map gams-olk-4 'gams-ol-select-item)

  (define-key map gams-olk-7 'gams-ol-mark)
  (define-key map "u" 'gams-ol-unmark)
  (define-key map "y" 'gams-ol-jump-mark)
  
  (define-key map "i" 'gams-ol-back-to-lst)
  (define-key map gams-olk-1 'gams-ol-help)
  (define-key map gams-olk-6 'gams-ol-view-quit)
  
  (define-key map "l" 'recenter)
  (define-key map [return] 'scroll-up)
  (define-key map [delete] 'scroll-down)
  (define-key map "1" 'delete-other-windows)
  (define-key map "n" 'gams-ol-next)
  (define-key map "p" 'gams-ol-prev)

  (define-key map "d" 'gams-lst-scroll-1)
  (define-key map "f" 'gams-lst-scroll-down-1)
  (define-key map "g" 'gams-lst-scroll-2)
  (define-key map "h" 'gams-lst-scroll-down-2)
  (define-key map "j" 'gams-lst-scroll-double)
  (define-key map "k" 'gams-lst-scroll-down-double)

  (define-key map "D" 'gams-lst-scroll-page-1)
  (define-key map "F" 'gams-lst-scroll-page-down-1)
  (define-key map "G" 'gams-lst-scroll-page-2)
  (define-key map "H" 'gams-lst-scroll-page-down-2)
  (define-key map "J" 'gams-lst-scroll-page-double)
  (define-key map "K" 'gams-lst-scroll-page-down-double)

  (define-key map "w" 'gams-lst-resize-frame)
  (define-key map "e" 'gams-lst-move-frame)
  )

;;; Menu for GAMS-OUTLINE mode.
(easy-menu-define 
  gams-ol-mode-menu gams-ol-mode-map "Menu keymap for GAMS-OUTLINE mode."
  '("GAMS-OUTLINE"
    ["Show the current item" gams-ol-view-base t]
    ["Show the next item" gams-ol-view-next t]
    ["Show the previous item" gams-ol-view-prev t]
    ["Select viewable items" gams-ol-select-item t]
    "--"
    ["Mark an item" gams-ol-mark t]
    ["Unmark an item" gams-ol-unmark t]
    ["Jump to the marked item" gams-ol-jump-mark t]
    "--"
    ["Switch back to the LST buffer" gams-ol-back-to-lst t]
    ["Show help" gams-ol-help t]
    ["Quit OUTLINE mode" gams-ol-view-quit t]
    "--"
    ["Widen the window" delete-other-windows t]
    ["Recentering" recenter t]
    ["Scroll up" scroll-up t]
    ["Scroll down" scroll-down t]
    ["Next line" gams-ol-next t]
    ["Previous line" gams-ol-prev t]
    ["Resize frame" gams-lst-resize-frame t]
    ["Move frame" gams-lst-move-frame t]
    ))


(defun gams-ol-mode (lst-file-buf)
"The GMS-OUTLINE mode.

\\[gams-ol-view-base]		Show the content of the item on the current line.
\\[gams-ol-view-next]		Show the content of the next item.
\\[gams-ol-view-prev]		Show the content of the previous item.
\\[gams-ol-select-item]		Select viewable items.

\\[gams-ol-mark]		Mark an item.
\\[gams-ol-unmark]		Unmark an item.
\\[gams-ol-jump-mark]		Jump to the marked line.

\\[gams-ol-back-to-lst]		Switch back to the LST buffer.
\\[gams-ol-help]		Show this help.
\\[gams-ol-view-quit]		Quit.

\\[delete-other-windows]		Widen the window.
\\[recenter]		Recenter.
\\[scroll-up] or RET	Scroll up the OUTLINE buffer.
\\[scroll-down] or DEL	Scroll down the OUTLINE buffer.
\\[delete-other-windows]		Widen the OUTLINE window.
\\[gams-ol-next]		Next line.
\\[gams-ol-prev]		Previous line.
\\[gams-lst-resize-frame]		Resize frame.
\\[gams-lst-move-frame]		Move frame.

Commands for Scrolling

Commands for scrolling are almost the same as the ones in the GAMS-LST mode.

Suppose that there are three windows displayed like

    __________________	  
   |   	      	      |  
   |  OUTLINE buffer |  ==>  OV-1
   |  CURSOR  here    |  
   |   	      	      |  
   |------------------|
   |   	      	      |  
   |  LST buffer 1    |  ==>  LST-1.
   |   	      	      |  
   |------------------|  
   |		      | 
   |  LST buffer 2    |  ==>  LST-2.
   |   	      	      |  
    ------------------

\\[gams-lst-scroll-1](\\[gams-lst-scroll-down-1])		Scroll the next buffer (LST-1) up (down) one line.
\\[gams-lst-scroll-2](\\[gams-lst-scroll-down-2])		Scroll the other next buffer (LST-2) up (down) one line.
\\[gams-lst-scroll-double](\\[gams-lst-scroll-down-double])		Scroll two LST buffers (LST-1 and LST-2) up (down) one line.

Keyboard.

  _____________________________________________________________
  |         |         |         |         |         |         |
  |    d    |    f    |    g    |    h    |    j    |    k    |
  |         |         |         |         |         |         |
  -------------------------------------------------------------

       |         |         |         |         |         |

      UP        DOWN      UP        DOWN      UP        DOWN
         LST-1               LST-2             LST-1 & 2


If only two window exists (e.g. OV-1 and LST-1),

\\[gams-lst-scroll-1] or '\\[gams-lst-scroll-2] or '\\[gams-lst-scroll-double]  ==> Scroll LST-1 up a line.
\\[gams-lst-scroll-down-1] or '\\[gams-lst-scroll-down-2] or '\\[gams-lst-scroll-down-double]  ==> Scroll LST-1 down a line.

The followings are page scroll commands.  Just changed to upper cases.

\\[gams-lst-scroll-page-1](\\[gams-lst-scroll-page-down-1])		Scroll the next buffer (LST-1) up (down) a page.
\\[gams-lst-scroll-page-2](\\[gams-lst-scroll-page-down-2])		Scroll the other next buffer (LST-2) up (down) a page.
\\[gams-lst-scroll-page-double](\\[gams-lst-scroll-page-down-double])		Scroll two buffers (LST-1 and LST-2) up (down) a page."
  (interactive)

  (setq major-mode 'gams-ol-mode)
  (setq mode-name "GAMS-OUTLINE")
  (use-local-map gams-ol-mode-map)

  (setq buffer-read-only t)

  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(gams-ol-font-lock-keywords t t))
  (font-lock-fontify-buffer)

  ;; Store the LST buffer name in a local variable.
  (make-local-variable 'gams-ol-lst)
  (setq gams-ol-lst lst-file-buf)

  ;; Create the local variable gams-ol-mark-flag.  gams-ol-mark-flag is
  ;; assigned non-nil if an item is marked in the OUTLINE buffer.
  (make-local-variable 'gams-ol-mark-flag)
  (setq gams-ol-mark-flag nil)

  ;; Run hook
  (run-hooks 'gams-ol-mode-hook)

  ;; Menu
  (easy-menu-add gams-ol-mode-menu)
)

(defun gams-ol-count-line ()
  "Calculate the current line number in the OUTLINE buffer."
  (let ((point-a (point))
	(count 1))
    (goto-char (point-min))
    (while
	(re-search-forward "[\n\C-m]" point-a t)
      (setq count (+ 1 count)))
    count))
    
(defun gams-ol-view-quit ()
  "Quit the GAMS-OUTLINE mode."
  (interactive)
  (let ((cur-buf (current-buffer)))
    ;; Unmark the marked item.
    (gams-ol-unmark)
    (if (buffer-live-p (get-buffer gams-ol-lst))
	(switch-to-buffer gams-ol-lst)
      (message "No correspoding LST buffer!")
      (sleep-for 1))
    (kill-buffer cur-buf)
    (delete-other-windows)))

(defun gams-ol-help ()
  "Show help in the GAMS-OUTLINE mode."
  (interactive)
  (describe-function 'gams-ol-mode))

(defun gams-ol-back-to-lst ()
  "Switch to the LST file buffer."
  (interactive)
  ;; Unmark.
  (gams-ol-unmark)
  (switch-to-buffer gams-ol-lst)
  (recenter))

(defun gams-ol-view-base ()
  "Show the content of the item on the current line in the another window."
  (interactive)
  (let ((line-num (gams-ol-count-line))
	list-par ele-par point-par
	(cur-buf (current-buffer))
	(flag gams-ol-mark-flag))
    (setq list-par (assoc line-num (gams-ol-get-alist t)))
    (setq ele-par (car list-par))
    (setq point-par (car (cdr list-par)))
    (if (null list-par)
	nil
      ;; If list-par is not nil.
      (if (one-window-p t)
	  ;; If only one window.
	  (if (null flag)
	      (progn
		(split-window
		 (selected-window)
		 (+ 1 gams-ol-height))
		(pop-to-buffer gams-ol-lst)
		(goto-char point-par)
		(recenter 1)
		(pop-to-buffer cur-buf))
	    (split-window
	     (selected-window)
	     (+ 1 gams-ol-height))
	    (pop-to-buffer gams-ol-lst)
	    (split-window
	     (selected-window)
	     (/ (window-height) 2))
	    (goto-char (car (cdr (assoc flag (gams-ol-get-alist)))))
	    (recenter 1)
	    (other-window 1)
	    (goto-char point-par)
	    (recenter 1)
	    (pop-to-buffer cur-buf))
	;; If two or more windows exits.
	(delete-other-windows)
	(if (null flag)
	    (progn
	      (split-window
	       (selected-window)
	       (+ 1 gams-ol-height))
	      (pop-to-buffer gams-ol-lst)
	      (goto-char point-par)
	      (recenter 1)
	      (pop-to-buffer cur-buf))
	  (split-window
	   (selected-window)
	   (+ 1 gams-ol-height-two))
	  (pop-to-buffer gams-ol-lst)
	  (split-window
	   (selected-window)
	   (/ (window-height) 2))
	  (goto-char (car (cdr (assoc flag (gams-ol-get-alist)))))
	  (recenter 1)
	  (other-window 1)
	  (goto-char point-par)
	  (recenter 1)
	  (pop-to-buffer cur-buf))
	);; If ends.
      (gams-ol-show-key)
      )))

(defun gams-ol-show-key ()
  "Show the basic keybindings in the GAMS-OUTLINE mode."
  (interactive)
  (message
   (format "%s = help, %s = show, [%s]oggle, [%s]ark, d,f,g,h,j,k = scroll, [%s]uit."
	   gams-olk-1
	   gams-olk-5
	   gams-olk-4
	   gams-olk-7
	   gams-olk-6)))

(defun gams-ol-next ()
  "Move to the next line."
  (interactive)
  (next-line 1)
  (gams-ol-show-key))

(defun gams-ol-prev ()
  "Move to the previous line."
  (interactive)
  (next-line -1)
  (gams-ol-show-key))

(defun gams-ol-view-next ()
  "Show the content of the item on the next line."
  (interactive)
  (next-line 1)
  (gams-ol-view-base))
  
(defun gams-ol-view-prev ()
  "Show the content of the item on the previous line."
  (interactive)
  (next-line -1)
  (gams-ol-view-base))

(defun gams-ol-mark ()
  "Mark an item."
  (interactive)
  (let ((cur-buff (current-buffer))
	(buffer-read-only nil)
	(line-num (gams-ol-count-line))
	(flag gams-ol-mark-flag))
    ;; Delete mark if flag is non-nil.
    (if flag
	(progn
	  (save-excursion
	    (goto-line flag)
	    (beginning-of-line)
	    (delete-char 1)
	    (insert " "))))
    (beginning-of-line)
    (delete-char 1)
    (insert "*")
    (beginning-of-line)
    (setq gams-ol-mark-flag line-num)
    ))

(defun gams-ol-count-win ()
  "Count the number of windows."
  (interactive)
  (let ((cur-win (selected-window))
	(num 1)
	flag)
    (save-excursion
      (if (one-window-p)
	  nil
	(catch 'quit
	  (while t
	    (other-window 1)
	    (setq flag (selected-window))
	    (cond
	     ((not (eq flag cur-win))
	      (setq num (+ num 1)))
	     ((eq flag cur-win)
	      (throw 'quit t)))))))
    num))
      
(defun gams-ol-unmark ()
  "Unmark the marked item."
  (interactive)
  (let
      ((buffer-read-only nil)
       line-num flag)
    (if gams-ol-mark-flag
	(save-excursion
	  (goto-char (point-min))
	  (if (re-search-forward "^* " nil t)
	      (progn
		(beginning-of-line)
		(delete-char 1)
		(insert " ")
		(setq gams-ol-mark-flag nil)))))))

(defun gams-ol-jump-mark()
  "Jumpt to the marked items."
  (interactive)
  (let ((line-num gams-ol-mark-flag))
    (if line-num
	(goto-line line-num))))

(defun gams-ol-make-alist ()
  "Make the alist of all items from the LST buffer.

The alist includes non-viewable items."
  (interactive)
  (let (point-a	point-b
	(par-index 0)			; Index of item.
	par-point			; Point of item.
	par-list
	flag-a
	(par-alist nil)
	(lst-buf (current-buffer)))
    (goto-char (point-min))
    ;;
    (while
	;; Search what can be items.
	(re-search-forward "^\\(----\\|               S O L V E      S U M M A R Y\\|\\*\\*\\*\\* REPORT SUMMARY\\|               L O O P S\\)" nil t)
      ;; Store the match.
      (setq flag-a (buffer-substring (match-beginning 1)
				     (match-end 1)))
      ;; If an item is found.
      (cond
       ;; The case for VAR, EQU, SET, PAR, COM.
       ((equal flag-a "----")
	(setq point-a (point))
	(forward-line 1)
	(setq point-b (point))
	(goto-char point-a)
	(cond
	 ;; set.
	 ((re-search-forward "[ ]+[0-9]*[ ]*\\(set\\)[ ]+\\([^ ]+\\)[ ]+\\(.*\\)" point-b t)
	  (beginning-of-line)
	  (setq par-point (point))
	  (end-of-line)

	  (setq par-index (+ 1 par-index))

	  (setq par-list (gams-ol-store-point))
	  (setq par-alist
		(cons
		 (append (list par-index par-point) par-list) par-alist)))
	 ;; parameter.
	 ((or
	   (re-search-forward "[ ]+[0-9]*[ ]*\\(parameter\\)[ ]+\\([^ ]+\\)[ ]+=[ ]+[0-9.]+[ ]+\\(.*\\)" point-b t)
	   (re-search-forward "[ ]+[0-9]*[ ]*\\(parameter\\)[ ]+\\([^ ]+\\)[ ]+\\(.*\\)" point-b t))
	  (beginning-of-line)
	  (setq par-point (point))
	  (end-of-line)
	  (setq par-index (+ 1 par-index))
	  (setq par-list (gams-ol-store-point))
	  ;; Make alist.
	  (setq par-alist
		(cons
		 (append (list par-index par-point) par-list) par-alist))
	  ;; Subprocess
	  (let (point-c point-e)
	    (while
		(progn
		  (setq point-c (point))
		  (re-search-forward "\\(\n\n\n\\|\f\\)" nil t)
		  (setq point-e (point))
		  (goto-char point-c)
		  (or (re-search-forward "^[ ]+\\(parameter\\)[ ]+\\([^ ]+\\)[ ]+=[ ]+[0-9.]+[ ]+\\(.*\\)" point-e t)
		      (re-search-forward "^[ ]+\\(parameter\\)[ ]+\\([^ ]+\\)[ ]*\\(.*\\)" point-e t)))
	      (beginning-of-line)
	      (setq par-point (point))
	      (end-of-line)
	      (setq par-index (+ 1 par-index))
	      (setq par-list (gams-ol-store-point))
	      (setq par-alist
		    (cons
		     (append (list par-index par-point) par-list) par-alist))
	      )))
	 ;; VARIABLE.
	 ((or
	   (re-search-forward "[ ]+[0-9]*[ ]*\\(variable\\)[ ]+\\([^ ]+\\)[ ]+=[ ]+[0-9.]+[ ]+\\(.*\\)" point-b t)
	   (re-search-forward "[ ]+[0-9]*[ ]*\\(variable\\)[ ]+\\([^ ]+\\)[ ]+\\(.*\\)" point-b t))
	  
	  (beginning-of-line)
	  (setq par-point (point))
	  (end-of-line)
	  (setq par-index (+ 1 par-index))
	  (setq par-list (gams-ol-store-point t))
	  (setq par-alist
		(cons
		 (append (list par-index par-point) par-list) par-alist))
	  ;; subprocess
	  (let (point-c point-e)
	    (while
		(progn
		  (setq point-c (point))
		  (re-search-forward "\\(\n\n\n\\|\f\\)" nil t)
		  (setq point-e (point))
		  (goto-char point-c)
		  (or (re-search-forward "^[ ]+\\(variable\\)[ ]+\\([^ ]+\\)[ ]+=[ ]+[0-9.]+[ ]+\\(.*\\)" point-e t)
		      (re-search-forward "^[ ]+\\(variable\\)[ ]+\\([^ ]+\\)[ ]*\\(.*\\)" point-e t)))
	      (beginning-of-line)
	      (setq par-point (point))
	      (end-of-line)
	      (setq par-index (+ 1 par-index))
	      (setq par-list (gams-ol-store-point t))
	      (setq par-alist
		    (cons
		     (append (list par-index par-point) par-list) par-alist)))))
	 ;; VAR.
	 ((re-search-forward "[ ]+\\(VAR\\)[ ]+\\([^ ]+\\)[ ]+\\(.*\\)" point-b t)
	  (beginning-of-line)
	  (setq par-point (point))
	  (end-of-line)
	  (setq par-index (+ 1 par-index))
	  (setq par-list (gams-ol-store-point))
	  (setq par-alist
		(cons
		 (append (list par-index par-point) par-list) par-alist)))
	 ;; EQU.
	 ((re-search-forward "[ ]+\\(EQU\\)[ ]+\\([^ ]+\\)[ ]+\\(.*\\)" point-b t)
	  (beginning-of-line)
	  (setq par-point (point))
	  (end-of-line)
	  (setq par-index (+ 1 par-index))
	  (setq par-list (gams-ol-store-point))
	  (setq par-alist
		(cons
		 (append (list par-index par-point) par-list) par-alist)))
	 ;; COM.
	 ((re-search-forward "[ ]+[0-9]+[ ]com:\\(.*\\)" point-b t)
	  (let (mark-a)
	    (setq mark-a (buffer-substring (match-beginning 1)
					   (match-end 1)))
	    (beginning-of-line)
	    (setq par-point (point))
	    (end-of-line)
	    (setq par-index (+ 1 par-index))
	    (setq par-alist
		  (cons
		   (list par-index par-point "COM" mark-a) par-alist))))
	 ;; OTHER.
	 ((re-search-forward "[ ]+[0-9]+[ ]\\(.*\\)" point-b t)
	  (let (mark-a)
	    (setq mark-a (buffer-substring (match-beginning 1)
					   (match-end 1)))
	    (beginning-of-line)
	    (setq par-point (point))
	    (end-of-line)
	    (setq par-index (+ 1 par-index))
	    (setq par-alist
		  (cons
		   (list par-index par-point "OTH" "" mark-a) par-alist))))))
       ;; REPORT SUMMARY
       ((equal flag-a "**** REPORT SUMMARY")
	(let (list-a)
	  (setq point-a (point))
	  (forward-line 1)
	  (setq point-b (point))
	  (goto-char point-a)
	  (end-of-line)
	  (setq par-index (+ 1 par-index))
	  (setq par-alist
		(cons
		 (list par-index point-a "SUM" "REPORT SUMMARY") par-alist))))
       ;; LOOP
       ((equal flag-a "               L O O P S")
	(let ((point-x (point))
	      point-y
	      par-type3 par-name3 par-exp3)
	  (end-of-line)
	  (setq point-y (point))
	  (goto-char point-x)
	  (while
	      (re-search-forward "[ ]+\\([^ ]+\\)[ ]+\\([^ ]+\\)" point-y t)
	    (setq  par-type3 "LOO")
	    (setq  par-name3 (buffer-substring
			      (match-beginning 1)
			      (match-end 1)))
	    (setq  par-exp3 (buffer-substring
			     (match-beginning 2)
			     (match-end 2)))
	    (save-excursion
	      (beginning-of-line)
	      (setq point-a (point)))
	    (setq par-index (+ 1 par-index))
	    (setq par-alist
		  (cons
		   (list par-index point-a par-type3 par-name3 par-exp3) par-alist))
	    ;; ?
	    (forward-line 1)
	    (setq point-x (point))
	    (end-of-line)
	    (setq point-y (point))
	    (goto-char point-x))))
       ;; Others.
       (t
	(let (list-a)
	  (setq point-a (point))
	  (forward-line 1)
	  (setq point-b (point))
	  (goto-char point-a)
	  (end-of-line)
	  (setq par-index (+ 1 par-index))
	  (setq par-alist
		  (cons
		   (list par-index point-a "SUM" "SOLVER SUMMARY"
			 (gams-ol-solve-sum))
		   par-alist))
	  ))))
    ;; 
    (setq gams-over-alist par-alist)
    ;; Return alist.
    gams-over-alist))

(defun gams-ol-solve-sum ()
  "Extract SOLVER STATUS and MODEL STATUS."
  (let (point-a point-b var-1 var-2 var-3)
    (save-excursion
      (setq point-a (point))
      (setq point-b (+ (point) 2000))
      (goto-char point-a)
      (re-search-forward "\*\*\*\* SOLVER STATUS[ ]+\\([0-9]*\\)\\(.*\\)[ ]*$" point-b t)
      (setq var-1 (buffer-substring (match-beginning 1)
				    (match-end 1)))
      (if (equal var-1 "")
	  (setq var-1 (buffer-substring (match-beginning 2)
					(match-end 2))))
      (re-search-forward "\*\*\*\* MODEL STATUS[ ]+\\([0-9]*\\)\\(.*\\)[ ]*$" point-b t)
      (setq var-2 (buffer-substring (match-beginning 1)
				    (match-end 1)))
      (if (equal var-2 "")
	  (setq var-2 (buffer-substring (match-beginning 2)
					(match-end 2))))
      ;; Remove the spaces in the line end.
      (cond
       (gams-emacs-21
	(setq var-1 (replace-regexp-in-string "[ ]+$" "" var-1))
	(setq var-2 (replace-regexp-in-string "[ ]+$" "" var-2)))
       (gams-emacs-20
	(setq var-1 (gams-replace-regexp-in-string "[ ]+$" "" var-1))
	(setq var-2 (gams-replace-regexp-in-string "[ ]+$" "" var-2))))
      (setq var-3 (concat "SOLVER STATUS = " var-1 ", " "MODEL STATUS = " var-2)))
    var-3))

;;; If Emacs 20, define replace-regexp-in-string.  This code is from
;;; subr.el in the Emacs 21 distribution.
(if gams-emacs-20
    (defun gams-replace-regexp-in-string (regexp rep string &optional
					    fixedcase literal subexp start)
      "Replace all matches for REGEXP with REP in STRING.

This code is from subr.el in Emacs 21 distribution.

Return a new string containing the replacements.

Optional arguments FIXEDCASE, LITERAL and SUBEXP are like the
arguments with the same names of function `replace-match'.  If START
is non-nil, start replacements at that index in STRING.

REP is either a string used as the NEWTEXT arg of `replace-match' or a
function.  If it is a function it is applied to each match to generate
the replacement passed to `replace-match'; the match-data at this
point are such that match 0 is the function's argument.

To replace only the first match (if any), make REGEXP match up to \\'
and replace a sub-expression, e.g.
  (replace-regexp-in-string \"\\(foo\\).*\\'\" \"bar\" \" foo foo\" nil nil 1)
    => \" bar foo\"
"

      ;; To avoid excessive consing from multiple matches in long strings,
      ;; don't just call `replace-match' continually.  Walk down the
      ;; string looking for matches of REGEXP and building up a (reversed)
      ;; list MATCHES.  This comprises segments of STRING which weren't
      ;; matched interspersed with replacements for segments that were.
      ;; [For a `large' number of replacments it's more efficient to
      ;; operate in a temporary buffer; we can't tell from the function's
      ;; args whether to choose the buffer-based implementation, though it
      ;; might be reasonable to do so for long enough STRING.]
      (let ((l (length string))
	    (start (or start 0))
	    matches str mb me)
	(save-match-data
	  (while (and (< start l) (string-match regexp string start))
	    (setq mb (match-beginning 0)
		  me (match-end 0))
       ;; If we matched the empty string, make sure we advance by one char
	    (when (= me mb) (setq me (min l (1+ mb))))
	    ;; Generate a replacement for the matched substring.
	    ;; Operate only on the substring to minimize string consing.
	    ;; Set up match data for the substring for replacement;
	    ;; presumably this is likely to be faster than munging the
	    ;; match data directly in Lisp.
	    (string-match regexp (setq str (substring string mb me)))
	    (setq matches
		  (cons (replace-match (if (stringp rep)
					   rep
					 (funcall rep (match-string 0 str)))
				       fixedcase literal str subexp)
			(cons (substring string start mb) ; unmatched prefix
			      matches)))
	    (setq start me))
	  ;; Reconstruct a string from the pieces.
	  (setq matches (cons (substring string start l) matches)) ; leftover
	  (apply #'concat (nreverse matches))))))


(defun gams-ol-convert (alist)
  "?"
  (interactive)
  (let ((temp-alist (reverse alist))
	(list-count 1)
	temp-element temp-alist-2 temp-list)
    (while temp-alist
      (setq temp-list (car temp-alist))
      (setq temp-element (car (cdr (cdr temp-list))))
      ;; If temp-element must be displayed or not?
      (if (cdr (assoc temp-element gams-ol-view-item))
	  ;; Display
	  (progn
	    (setcar temp-list list-count)
	    (setq temp-alist-2
		  (cons temp-list temp-alist-2))
	    (setq list-count (+ 1 list-count)))
	;; Not display.
	nil)
      (setq temp-alist (cdr temp-alist)))
    (setq gams-ol-alist-tempo temp-alist-2))
  gams-ol-alist-tempo)

(defun gams-ol-show (buffname alist)
  "?.

BUFFNAME is the OV buffer name.
ALIST is the alist of all items."
  (interactive)
  (let ((temp-alist (reverse alist))
	(lst-buf (current-buffer))
	list-list
	point-end)
    (switch-to-buffer buffname)
    (setq buffer-read-only nil)
    (setq truncate-lines t)
    (erase-buffer)
    (goto-char (point-min))
    ;; Insert.
    (if (null temp-alist)
	(insert "No viewable item in GAMS-LST-OUTLINE mode!  Type `q' to quit or type `t' to toggle!")
      (while temp-alist
	(goto-char (point-max))
	(setq list-list (car temp-alist))
	(gams-lst-insert-item list-list)
	(setq temp-alist (cdr temp-alist)))
      (goto-char (point-min)))))

(defun gams-lst-insert-item (list-list)
  "Insert item."
  (interactive)
  (let ((list-num (car list-list))
	(list-po (car (cdr list-list)))
	(list-type (car (cdr (cdr list-list))))
	(list-name (car (cdr (cdr (cdr list-list)))))
	(list-cont (car (cdr (cdr (cdr (cdr list-list)))))))
    (if (null (equal list-type "COM"))
	;; Not COM.
	(progn
	  (move-to-column 2 t)
	  (insert list-type)
	  (insert "  ")
	  (insert list-name)
	  (if list-cont
	      (progn
		(move-to-column 26 t)
		(insert list-cont))))
      ;; COM.
      (insert (concat "COM:" " " list-name)))
    (insert "\n")))
  
(setq gams-ol-item-buffer "*GAMS-SELECT-ITEM*")

(defun gams-ol-select-item ()
  "Evoke the SELECT-ITEM mode."
  (interactive)
  (let ((cur-buf (current-buffer))
	(temp-buf gams-ol-item-buffer)
	(item-list gams-ol-item)
	temp-item
	(temp-num 1)
	(flag nil))
    (pop-to-buffer (get-buffer-create temp-buf))
    (erase-buffer)
    (goto-char (point-min))
    (while item-list
      (setq temp-item (car item-list))
      (if (gams-ol-check-included temp-item t)
	  (setq flag t)
	(setq flag nil))
      (goto-char (point-max))
      (insert
       (concat " " " ["
	       (if flag "+" "-")
	       "] " temp-item "\n"))
      (setq item-list (cdr item-list))
      (setq temp-num (+ temp-num 1)))
    (goto-char (point-max))
    (gams-ol-select-key)
    (goto-char (point-min))
    (gams-ol-select-mode cur-buf)))

;;; key assignment for SELECT-ITEM mode.
(defvar gams-os-on "j")
(defvar gams-os-on-prev "i")
(defvar gams-os-off "h")
(defvar gams-os-off-prev "u")
(defvar gams-os-on-all "d")
(defvar gams-os-off-all "f")
(defvar gams-os-quit "\r")
(defvar gams-os-sum "s")
(defvar gams-os-var "v")
(defvar gams-os-equ "e")
(defvar gams-os-vri "r")
(defvar gams-os-set "t")
(defvar gams-os-par "p")
(defvar gams-os-loo "l")
(defvar gams-os-oth "o")
(defvar gams-os-com "c")

(defvar gams-ol-select-mode-map (make-keymap) "Keymap.")
(let ((map gams-ol-select-mode-map)
      (temp-list (reverse gams-ol-item))
      temp-name
      (count 9))
  (define-key map " " 'gams-ol-toggle)
  (define-key map gams-os-on 'gams-ol-toggle-on)
  (define-key map gams-os-off 'gams-ol-toggle-off)
  (define-key map gams-os-on-prev 'gams-ol-toggle-on-prev)
  (define-key map gams-os-off-prev 'gams-ol-toggle-off-prev)
  (define-key map gams-os-quit 'gams-ol-select-quit)
  (define-key map gams-os-on-all 'gams-ol-toggle-all-on)
  (define-key map gams-os-off-all 'gams-ol-toggle-all-off)

  (define-key map gams-os-var 'gams-ol-toggle-var)
  (define-key map gams-os-equ 'gams-ol-toggle-equ)
  (define-key map gams-os-vri 'gams-ol-toggle-vri)
  (define-key map gams-os-set 'gams-ol-toggle-set)
  (define-key map gams-os-par 'gams-ol-toggle-par)
  (define-key map gams-os-sum 'gams-ol-toggle-sum)
  (define-key map gams-os-loo 'gams-ol-toggle-loo)
  (define-key map gams-os-oth 'gams-ol-toggle-oth)
  (define-key map gams-os-com 'gams-ol-toggle-com)
  )

;;; Menu for OUTLINE-SELECT mode.
(easy-menu-define 
  gams-ol-select-mode-menu gams-ol-select-mode-map "Menu keymap for OUTLINE-SELECT mode."
  '("OUTLINE-SELECT"
    ["Toggle" gams-ol-toggle t]
    ["Toggle on (next)" gams-ol-toggle-on t]
    ["Toggle off (next)" gams-ol-toggle-off t]
    ["Toggle on (previous)" gams-ol-toggle-on-prev t]
    ["Toggle off (previous)" gams-ol-toggle-off-prev t]
    ["Toggle on all" gams-ol-toggle-all-on t]
    ["Toggle off all" gams-ol-toggle-all-off t]
    "--"
    ["Quit SELECT mode" gams-ol-select-quit t]
    "--"
    ["Toggle VAR" gams-ol-toggle-var t]
    ["Toggle EQU" gams-ol-toggle-equ t]
    ["Toggle VRI" gams-ol-toggle-vri t]
    ["Toggle SET" gams-ol-toggle-set t]
    ["Toggle PAR" gams-ol-toggle-par t]
    ["Toggle SUM" gams-ol-toggle-sum t]
    ["Toggle LOO" gams-ol-toggle-loo t]
    ["Toggle OTH" gams-ol-toggle-oth t]
    ["Toggle COM" gams-ol-toggle-com t]
    ))

(defun gams-ol-select-key ()
    (insert (concat "\n"
		    "Key:\n"
		    (format "[SPC]    = Toggle\n")
		    (format "[%s/%s]    = Toggle on\n"
			    gams-os-on gams-os-on-prev)
		    (format "[%s/%s]    = Toggle off\n"
			    gams-os-off gams-os-off-prev)
		    (format "[%s]      = Toggle on all\n" gams-os-on-all)
		    (format "[%s]      = Toggle off all\n" gams-os-off-all)
		    (format "[RET]    = Quit.\n\n")
		    (format "[%s] = SUM, " gams-os-sum)
		    (format "[%s] = VAR, " gams-os-var)
		    (format "[%s] = EQU, " gams-os-equ)
		    (format "[%s] = PAR, " gams-os-par)
		    (format "[%s] = SET,\n" gams-os-set)
		    (format "[%s] = VRI, " gams-os-vri)
		    (format "[%s] = LOO, " gams-os-loo)
		    (format "[%s] = OTH, " gams-os-oth)
		    (format "[%s] = COM." gams-os-com))))

(defun gams-ol-select-mode (buffname)
  "Start the SELECT-ITEM mode.

BUFFNAME is the OUTLINE buffer name."
  (interactive)
  (setq major-mode 'gams-ol-select-mode)
  (setq mode-name "GAMS-SELECT-ITEM")
  (use-local-map gams-ol-select-mode-map)
  (setq buffer-read-only t)
;  (make-local-variable 'font-lock-defaults)
;  (setq font-lock-defaults '(gams-ol-font-lock-keywords t t))
;  (font-lock-fontify-buffer)
  (make-local-variable 'gams-ol-olbuff)
  (setq gams-ol-olbuff buffname)
  ;; Make local variable.  `gams-ol-item-flag' is a buffer local
  ;; variable. When the SELECT buffer is created, `gams-ol-item-flag' is
  ;; given the same content as `gams-ol-view-item'.
  ;; `gams-ol-item-flag' may be modified but `gams-ol-view-item'
  ;; reserve its initial value.
  (make-local-variable 'gams-ol-item-flag)
  (setq gams-ol-item-flag
	(gams-ol-check-func gams-ol-view-item))

  ;;
;  (easy-menu-add gams-ol-select-mode-menu)
  )
;; gams-ol-select-mode ends here.

(defun gams-ol-check-func (alist)
  "Copy the ALIST to the new alist."
  (interactive)
  (let ((temp-alist (reverse alist))
	temp-list
	temp-item
	res-alist)
    (while temp-alist
      (setq temp-item (car temp-alist))
      (if (cdr temp-item)
	  (setq res-alist
		(append
		 (list (cons (car temp-item) t)) res-alist))
	(setq res-alist
	      (append
	       (list (cons (car temp-item) nil)) res-alist)))
      (setq temp-alist (cdr temp-alist)))
    res-alist))

(defun gams-ol-select-judge ()
  "Judge the item on the line and return its value."
  (let (point-end temp-item)
    (save-excursion
      (end-of-line)
      (setq point-end (point))
      (beginning-of-line)
      (if (re-search-forward
	   "\\([+]\\|[-]\\)[]][ ]+\\([A-Z][A-Z][A-Z]\\)$" point-end t)
	  (setq temp-item (buffer-substring
			   (match-beginning 2)
			   (match-end 2))))
    temp-item)))

(defun gams-ol-toggle (&optional on off prev)
  "Toggle check.

If on in non-nil, toggle on. If off is non-nil, toggle off.
If prev is non-nil, move up after toggle."
  (interactive)
  (let* ((buffer-read-only nil)
	(item (gams-ol-select-judge))
	(flag (gams-ol-check-included item)))
    (if (null item)
	;; if no item on the current line, Do nothing.
	nil
      ;; if any item on the current line.
      (beginning-of-line)
      (move-to-column 3)
      (cond
       ;; Just toggle.
       ((and (null on) (null off))
	;; Delete.
	(delete-char 1)
	(if flag
	    ;; If checked.
	    (insert "-")
	  ;; Not checked.
	  (insert "+"))
	(gams-ol-check-toggle item))
       ;; Toggle on.
       ((and on (null off))
	(if flag
	    nil
	  ;; Delete.
	  (delete-char 1)
	  (insert "+")
	  (gams-ol-check-toggle item)))
       ;; Toggle off.
       ((and (null on) off)
	(if (null flag)
	    nil
	  ;; Delete.
	  (delete-char 1)
	  (insert "-")
	  (gams-ol-check-toggle item)))))
    ;; forward or backward?
    (if prev
	(forward-line -1)
      (forward-line 1))
    ))

(defun gams-ol-toggle-on ()
  "Toggle on the item on the current line."
  (interactive)
  (gams-ol-toggle t))

(defun gams-ol-toggle-on-prev ()
  "Toggle on the item on the current line."
  (interactive)
  (gams-ol-toggle t nil t))

(defun gams-ol-toggle-all-on ()
  "Toggle on all items."
  (interactive)
  (let ((times 9))
    (save-excursion
      (goto-char (point-min))
      (while (> times 0)
	(gams-ol-toggle t)
	(setq times (- times 1))))))

(defun gams-ol-toggle-all-off ()
  "Toggle off all the items."
  (interactive)
  (let ((times 9))
    (save-excursion
      (goto-char (point-min))
      (while (> times 0)
	(gams-ol-toggle nil t)
	(setq times (- times 1))))))

(defun gams-ol-toggle-off ()
  "Toggle off the item on the current line."
  (interactive)
  (gams-ol-toggle nil t))

(defun gams-ol-toggle-off-prev ()
  "Toggle off the item on the current line."
  (interactive)
  (gams-ol-toggle nil t t))

(defun gams-ol-toggle-func (item)
  "item is an item name."
  (goto-char (point-min))
  (goto-char (re-search-forward (concat "] " item) nil t))
  (beginning-of-line)
  (gams-ol-toggle))

(defun gams-ol-toggle-var ()
  (interactive)
  (gams-ol-toggle-func "VAR"))

(defun gams-ol-toggle-equ ()
  (interactive)
  (gams-ol-toggle-func "EQU"))

(defun gams-ol-toggle-vri ()
  (interactive)
  (gams-ol-toggle-func "VRI"))

(defun gams-ol-toggle-par ()
  (interactive)
  (gams-ol-toggle-func "PAR"))

(defun gams-ol-toggle-set ()
  (interactive)
  (gams-ol-toggle-func "SET"))

(defun gams-ol-toggle-loo ()
  (interactive)
  (gams-ol-toggle-func "LOO"))

(defun gams-ol-toggle-sum ()
  (interactive)
  (gams-ol-toggle-func "SUM"))

(defun gams-ol-toggle-oth ()
  (interactive)
  (gams-ol-toggle-func "OTH"))

(defun gams-ol-toggle-com ()
  (interactive)
  (gams-ol-toggle-func "COM"))

(defun gams-ol-select-quit ()
  "Quit the SELECT-ITEM mode."
  (interactive)
  (let ((cur-buff (current-buffer))
	(ov-buff gams-ol-olbuff)
	(item gams-ol-item-flag)
	temp-buf)
    (if (buffer-live-p ov-buff)
	;; If OUTLINE buffer exits.
	(progn
	  (switch-to-buffer ov-buff)
	  (kill-buffer cur-buff)
	  (delete-other-windows)
	  (if (equal item gams-ol-view-item)
	      ;; If no change has been made to gams-ol-view-item.
	      nil
	    ;; If any change has been made.
	    ;; Switch to the LST buffer.
	    (switch-to-buffer gams-ol-lst)
	    ;; Store the lst buffer name.
	    (setq temp-buf (current-buffer))
	    (setq gams-ol-view-item item)
	    (gams-ol-show ov-buff (gams-ol-convert (gams-ol-get-alist)))
	    (switch-to-buffer ov-buff)
	    (setq gams-ol-mark-flag nil)
	    ;; give lst buffer name.
	    (gams-ol-mode temp-buf))
	  (gams-ol-show-key))
      ;; If OUTLINE buffer does not exists.
      (message "No OUTLINE buffer exists!")
      (sleep-for 1)
      (kill-buffer cur-buff))))


(defun gams-ol-check-included (item &optional flag)
  "Judge whether the item is checked or not.
If the flag is non-nil, use `gams-ol-view-item'.
"
  (let ((temp-alist
	 (if flag gams-ol-view-item gams-ol-item-flag)))
    (if (cdr (assoc item temp-alist))
	t
      nil)))

(defun gams-ol-check-toggle (item)
  "Toggle the check of the item."
  (let ((temp-alist gams-ol-item-flag))
    (if (gams-ol-check-included item)
	;; If checked.
	(setcdr (assoc item temp-alist) nil)
      ;; If not checked.
      (setcdr (assoc item temp-alist) t))
    gams-ol-view-item
    gams-ol-item-flag))


;;; provide.

(provide 'gams)

;;(load "gams-new.el")

;;; GAMS.EL ends here
;Local variables:
;mode: emacs-lisp
;syntax: elisp
;End:
